#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::ContextManager;
use strict;

=head1 ContextManager

ContextManager is actually a bit of a misnomer, due to the misnomer that is the
QueryContext class.  The ContextManager class and its descendents are
responsible for managing each context and access to it.  Each ContextManager
holds a set of objects that comprises its data-set.  Because we may and do need
to decorate accesses to the objects, all requests to look at fields and the
like go through us.  For example, it provides us with the ability to introduce
synthetic attributes dynamically.  It's also fairly essential for the set
semantics; it would suck for the caller to have to iterate over a set of objects
to accumulate a scalar value from each of them into a set.  Instead, we do that
for them.

=cut

sub new
{
    my $class = shift;
    my $self  =
    {
        NODES => [],
        PARENT => undef,
        OWNER => undef,
        HINTS => {},
    };
    
    if(@_)
    {
        $self->{NODES} = shift;
    }
    
    bless ($self, $class);
    return $self;
}

my $forbiddenEdges = 
    {
        'MT::Author' => {'password' => 1, 'hint' => 1},
    };

sub nodes
{
    my $self = shift;
    
    if(@_)
    {
        $self->{NODES} = shift;
    }
    
    return $self->{NODES};
}

sub addNode
{
    my $self = shift;
    my $newNode = shift;
    
    push(@{$self->{NODES}}, $newNode)
}

sub parent
{
    my $self = shift;
    
    if(@_)
    {
        $self->{PARENT} = shift;
    }
    
    return $self->{PARENT};
}

sub owner
{
    my $self = shift;
    
    if(@_)
    {
        $self->{OWNER} = shift;
    }
    
    return $self->{OWNER};
}

sub hint
{
    my $self = shift;
    my $key = shift;
    
    if(@_)
    {
        $self->{HINTS}->{$key} = shift;
    }

    return $self->{HINTS}->{$key};
}

sub findHint
{
    my $self = shift;
    my $key = shift;
    
    my $val;
    
    $val = $self->hint($key);
    if(!defined($val) && defined($self->{OWNER}))
    {
        $val = $self->{OWNER}->findHint($key);
    }
    if(!defined($val) && defined($self->{PARENT}))
    {
        $val = $self->{PARENT}->findHint($key);
    }
    
    return $val;
}

sub loop
{
    my $self = shift;
    my $func = shift;

    my $curNodes = $self->nodes();
    
    my $loopContext = OnlyTheWind::KoalaRainbow::ContextManager->new();
    $loopContext->owner($self);
    $loopContext->hint('count', scalar @$curNodes);
       
    my $position = 0;
    for my $node (@$curNodes)
    {
        $loopContext->nodes([$node]);
        $loopContext->hint('position', $position);

        $func->($loopContext, $node);	
	
        $position++;
    }
    
    1;
}

################################################################################
# The following function is from Richard Clamp's UNIVERSAL::canAUTOLOAD package.
#  It is Copyright (C) 2003 Richard Clamp, and licensed under the "same terms as
#  Perl itself."
#
#  The full author credit is:
#  "Richard Clamp <richardc@unixbeard.net> original need and anticipated
#   documentation from Mark Fowler."
#
#  The full copyright credit is:
#  "Copyright (C) 2003 Richard Clamp. All Rights Reserved.
#   This module is free software; you can redistribute it and/or modify it under
#   the same terms as Perl itself."
#
#  This method is embedded to eliminate the need for a separate package to be
#   installed to this end.  I believe this satisfies the requirements of the
#   artistic license (thereby compliant with sixapart usage), especially
#   as this method is private and may be eligible for the 'commercial'
#   aggregation clause.  Regardless, the source is reproduced with only a
#   trivial modification and all copyright notices and associated disclaimers
#   are reproduced.  Likewise, the source is available here for your perusal,
#   so I believe this to be fully license compliant.
sub _can {
    use Class::ISA;
    no warnings 'redefine';

    my ($referent, $want) = @_;

    my $class = ref $referent || $referent;
    my @path = ( Class::ISA::self_and_super_path( $class ), 'UNIVERSAL' );

    # first look for a solid method
    for my $search (@path) {
        return \&{"$search\::$want"} if exists &{"$search\::$want"};
    }

    # then look for an AUTOLOAD sub
    for my $search (@path) {
        next unless exists &{"$search\::AUTOLOAD"};
        my $code = "package $search;".
          'sub { our $AUTOLOAD = "$class\::$want"; goto &AUTOLOAD }';
        my $sub = eval $code or die "compiling '$code': $@";
        return $sub;
    }

    # no? give up
    return undef;
}

# End Richard Clamp UNIVERSAL::canAUTOLOAD::Can
###############################################

sub evalEdge
{
    my $self = shift;
    my ($edgeName) = @_;
    
    my $nodes = $self->{NODES};
    
    my $outset = [];
    
    for my $node (@$nodes)
    {
        # make sure the edge name is permitted:
        my $nodeClass = ref($node);
        if(exists $forbiddenEdges->{$nodeClass})
        {
            if($forbiddenEdges->{$nodeClass}->{$edgeName})
            {
                die "Edge '$edgeName' is forbidden for class '$nodeClass'";
            }
        }
    
        my $meth = _can($node, $edgeName);
        if($meth)
        {
            # call the method.
            # What could the method return, and how do we deal with it?
            # - object (reference): it's good as is.
            # - array ref: we need to dump its contents out, I think.
            # - scalar: okay
            # we just can't handle pure arrays/hashes, but I think we're okay
            my $outnode = $meth->($node);
            
            
            if($outnode)
            {
                push(@$outset, $outnode);
            }
        }
        else
        {
            # error!
            die 'Node ' . $node . ' has no method "' . $edgeName . '"';
        }

    }
    
    my $resultCtxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new($outset);
    return $resultCtxMgr;
}

1;
