#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::DrawEngines::ImageMagick;
use strict;

use Image::Magick;

sub new
{
    my $class = shift;
    my $self  =
    {
        FILENAME => undef,
        IMAGE => undef,
        LENSES => [],
    };
    
    bless ($self, $class);
    return $self;
}

sub init
{
    my $self = shift;
    my %args = @_;
    
    my $image = Image::Magick->new(size => $args{width} . 'x' . $args{height},
                                  );
                                  
    $self->{IMAGE} = $image;
    
    if($args{background})
    {
        $image->ReadImage('xc:' . $self->_color($args{background}));
    }
    else
    {
        $image->ReadImage('xc:white');
    }
}

sub lenses
{
    my $self = shift;
    
    if(@_)
    {
        $self->{LENSES} = shift;
    }
    
    return $self->{LENSES};
}

sub _color
{
    my $self = shift;
    my ($colorNode) = @_;
    
    if(!defined($colorNode))
    {
        die 'Received undefined color.';
    }
    if(ref($colorNode) ne 'OnlyTheWind::KoalaRainbow::DrawNodes::Color')
    {
        die "Expecting color node, got: '" . ref($colorNode) . "'";
    }
    
    if($colorNode->name)
    {
        return $colorNode->name;
    }
    elsif($colorNode->hexcode)
    {
        return $colorNode->hexcode;
    }
    else
    {
        my ($r, $g, $b) = $colorNode->rgb;
        return 'rgb(' . $r . ', ' . $g . ', ' . $b . ')';
    }
}

sub _applyLenses
{
    my $self = shift;
    my $lenses = $self->{LENSES};
    my @transformed = ();
    
    while(@_)
    {
        my $x = shift;
        my $y = shift;
        
        for my $lensFunc (@$lenses)
        {
            ($x, $y) = $lensFunc->($x, $y);
        }
        
        push(@transformed, $x);
        push(@transformed, $y);
    }
    
    return @transformed;
}

sub rect
{
    my $self = shift;    
    my ($left, $top, $right, $bottom, $fillColor, $borderColor) = @_;

    ($left, $top, $right, $bottom) =
        $self->_applyLenses($left, $top, $right, $bottom);

    my $image = $self->{IMAGE};

    $image->Draw(primitive => 'rectangle',
                 points => $left . ',' . $top . ' ' . $right . ',' . $bottom,
                 fill => $self->_color($fillColor),
                 stroke => $self->_color($borderColor),
                );
}

sub circle
{
    my $self = shift;    
    my ($cx, $cy, $perimx, $perimy, $fillColor, $borderColor) = @_;

    ($cx, $cy, $perimx, $perimy) =
        $self->_applyLenses($cx, $cy, $perimx, $perimy);

    my $image = $self->{IMAGE};

    $image->Draw(primitive => 'circle',
                 points => $cx . ',' . $cy . ' ' . $perimx . ',' . $perimy,
                 fill => $self->_color($fillColor),
                 stroke => $self->_color($borderColor),
                );
}

sub ellipse
{
    my $self = shift;    
    my ($left, $top, $right, $bottom, $fillColor, $borderColor,
        $startAng, $endAng) = @_;

    ($left, $top, $right, $bottom) =
        $self->_applyLenses($left, $top, $right, $bottom);

    my $rx = ($right - $left) / 2;
    my $ry = ($bottom - $top) / 2;
    
    my $cx = $left + $rx;
    my $cy = $top + $ry;

    if(!defined($startAng))
    {
        $startAng=0;
    }
    if(!defined($endAng))
    {
        $endAng=360;
    }

    my $image = $self->{IMAGE};

    # use the normal ellipse case if no fancy angle is involved.
    if($startAng == 0 and $endAng == 360)
    {
        $image->Draw(primitive => 'ellipse',
                     points => $cx . ',' . $cy . ' ' . $rx . ',' . $ry .
                               ' ' . $startAng . ',' . $endAng,
                     fill => $self->_color($fillColor),
                     stroke => $self->_color($borderColor),
                    );
        return;
    }
    
    # I'm confident pi is a constant somewhere
    my $srad = $startAng / 180 * 3.1415926;
    my $erad = $endAng / 180 * 3.1415926;
    
    my $largeFlag;
    my $sweepFlag;
    if(($endAng - $startAng) > 180)
    {
        $largeFlag = 1;
    }
    else
    {
        $largeFlag = 0;        
    }
    
    $sweepFlag = 1;
    
    my $sx = cos($srad) * $rx;
    my $sy = sin($srad) * $ry;
    
    my $ex = cos($erad) * $rx;
    my $ey = sin($erad) * $ry;

    my $rex = $ex - $sx;
    my $rey = $ey - $sy;

    # NOTE: if the angle subtended is <= 45 degrees, the result is that the
    #  arc is not drawn.  I am currently blaming this on ImageMagick on the
    #  basis that independently varying the start and stop angles seems to
    #  correlate with the angle subtended rather than a specific start or end
    #  angle which might then correspond to a truncation error on my part
    #  somehow creating an inconsistent command.  Likewise, just flipping the
    #  sweepFlag to off and then tweaking values seems quite happy, suggesting
    #  that my data passes (working) data validation.
   
    my $cmd = '';
    # Move to 
    $cmd .= 'M' . $cx . ',' . $cy . ' ';
    $cmd .= 'l ' . int($sx) . ',' . int($sy) . ' ';
#    $cmd .= 'a' . $rx . ',' . $ry . ' 0 1,0 ';
    $cmd .= 'a' . $rx . ',' . $ry . ' 0 ' . $largeFlag . ',' . $sweepFlag . ' ';
    $cmd .= int($rex) . ',' . int($rey) . ' ';
    $cmd .= 'z';

#warn $cmd;

    $image->Draw(primitive => 'path',
                 points => $cmd,
                 fill => $self->_color($fillColor),
                 stroke => $self->_color($borderColor),
                );
                 
    
}

sub line
{
    my $self = shift;    
    my ($x1, $y1, $x2, $y2, $color, $width) = @_;

    ($x1, $y1, $x2, $y2) = $self->_applyLenses($x1, $y1, $x2, $y2);

    my %optargs = ();
    if($width)
    {
        $optargs{strokewidth} = $width;
    }

    # fill gets you the default 1-pixel wide.  to have it use linewidth,
    #  must use stroke (and therefore it for color)!
    my $image = $self->{IMAGE};
    $image->Draw(primitive => 'line',
                 points => $x1 . ',' . $y1 . ' ' . $x2 . ',' . $y2,
                 stroke => $self->_color($color),
                 %optargs
                );
}

sub filter
{
    my $self = shift;
    my ($filterArgs) = @_;
    
    my $filterName = $filterArgs->{name};
    
    my $image = $self->{IMAGE};
       
    if($filterName eq 'blur')
    {
        $image->Blur(radius => (exists $filterArgs->{radius}) ?
                                $filterArgs->{radius} : 0.0,
                     sigma => (exists $filterArgs->{sigma}) ? 
                              $filterArgs->{sigma} : 5.0,
                     channel => 'all');
    }
    elsif($filterName eq 'shade')
    {
        my $gray;
    
        if(exists $filterArgs->{gray})
        {
            $gray = $filterArgs->{gray};
            if(($gray ne 'true') and ($gray ne 'false'))
            {
                if($gray)
                {
                    $gray = 'true';
                }
                else
                {
                    $gray = 'false';
                }
            }
        }
        else
        {
            $gray = 'false';
        }
    
        $image->Shade(azimuth => (exists $filterArgs->{azimuth}) ?
                                $filterArgs->{azimuth} : 45.0,
                      elevation => (exists $filterArgs->{elevation}) ? 
                              $filterArgs->{elevation} : 120.0,
                      gray => $gray,
                      );        
    }
}


sub renderToBlob
{
    my $self = shift;    
    my $image = $self->{IMAGE};
    
    return $image->ImageToBlob(magick => 'png');
}

1;
