#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::Drawing;
use strict;

use OnlyTheWind::KoalaRainbow::DrawEngines::EngineFactory;

use OnlyTheWind::KoalaRainbow::DrawEngines::ImageMap;

sub new
{
    my $class = shift;
    my $self  =
    {
        NODES  => [],
        BOUNDS => [0, 0, 0, 0],
        LAYERS => {},
        AUTO_SCALE => 0,
        SCALE_LAYER => undef,
        WIDTH => undef,
        HEIGHT => undef,
        FILTERS => {},
        PADDING => [0, 0, 0, 0],
        BACKGROUND => undef,
        IMAGE_MAP_NAME => undef,
        IMAGE_MAP_CONTENTS => '',
    };
    
    bless ($self, $class);
    return $self;
}

sub width
{
    my $self = shift;
    
    if (@_)
    {
        $self->{WIDTH} = shift;
    }
    
    return $self->{WIDTH};
}

sub height
{
    my $self = shift;
    
    if (@_)
    {
        $self->{HEIGHT} = shift;
    }
    
    return $self->{HEIGHT};
}

sub background
{
    my $self = shift;
    
    if (@_)
    {
        $self->{BACKGROUND} = shift;
    }
    
    return $self->{BACKGROUND};
}

sub autoScale
{
    my $self = shift;
    
    if (@_)
    {
        $self->{AUTO_SCALE} = shift;
    }
    
    return $self->{AUTO_SCALE};
}

sub scaleLayer
{
    my $self = shift;
    
    if (@_)
    {
        $self->{SCALE_LAYER} = shift;
    }
    
    return $self->{SCALE_LAYER};
}

sub imageMapName
{
    my $self = shift;
    
    if (@_)
    {
        $self->{IMAGE_MAP_NAME} = shift;
    }
    
    return $self->{IMAGE_MAP_NAME};
}

sub imageMapContents
{
    my $self = shift;
    
    if (@_)
    {
        $self->{IMAGE_MAP_CONTENTS} = shift;
    }
    
    return $self->{IMAGE_MAP_CONTENTS};
}

sub pad
{
    my $self = shift;

    if(@_)
    {
        $self->{PADDING} = \@_; # I hope this is safe.
    }
    
    return $self->{PADDING};
}

my $supportedFilters = {'blur' => 1, 'shade' => 1,};

sub supportsFilter
{
    my $self = shift;
    my $filterName = shift;
    
    return exists $supportedFilters->{$filterName};
}

sub calculateBounds
{
    my $self = shift;
    my ($left, $top, $right, $bottom) = (undef, undef, undef, undef);

    my $scalelayer = $self->{SCALE_LAYER};

    my $nodes = $self->{NODES};
    for my $node (@$nodes)
    {
        if(defined($scalelayer))
        {
            if($node->layer != $scalelayer)
            {
                next;
            }
        }
    
        my ($nl, $nt, $nr, $nb) = $node->bounds();

        # let's try this crazy perl magic syntax
        $left = $nl if !defined($left) or ($nl < $left);
        $top  = $nt if !defined($top) or ($nt < $top);
        $right  = $nr if !defined($right) or ($nr > $right);
        $bottom = $nb if !defined($bottom) or ($nb > $bottom);
        # no sir, I don't like it.
        
        #warn "boogedy: $left $top $right $bottom";
    }
    
    $self->{BOUNDS} = [$left - 1, $top - 1, $right + 1, $bottom + 1];
    
    return $self->{BOUNDS};
}

sub calculateLayers
{
    my $self = shift;
    my $layers = {};

    my $nodes = $self->{NODES};
    for my $node (@$nodes)
    {
        my $nl = $node->layer();
        
        my $nodesInLayer;
        
        if(exists $layers->{$nl})
        {
            $nodesInLayer = $layers->{$nl};
            push(@$nodesInLayer, $node);
        }
        else
        {
            $nodesInLayer = [$node];
            $layers->{$nl} = $nodesInLayer;
        }

    }
    
    $self->{LAYERS} = $layers;
    
    return $layers;
}

sub addNode
{
    my $self = shift;
    my ($newNode) = @_;
    
    push(@{$self->{NODES}}, $newNode);
    return $newNode;
}

sub addFilter
{
    my $self = shift;
    my ($filterArgs) = @_;
    
    my $layer = $filterArgs->{layer};
    
    my $filters = $self->{FILTERS};
    if(exists $filters->{$layer})
    {
        push(@{$filters->{$layer}}, $filterArgs);
    }
    else
    {
        $filters->{$layer} = [$filterArgs];
    }
    
    1;
}

sub _makeLenses
{
    my $self = shift;
    my $lenses = [];
    
    if($self->{AUTO_SCALE})
    {
        #warn 'REF: ' . ref($self->{BOUNDS});
        my ($bx, $by, $br, $bb) = @{$self->{BOUNDS}};
        
        # handle the zero-width, zero-height cases for increased happiness.
        if($bx == $br)
        {
            $bx -= 1;
            $br += 1;
        }
        if($by == $bb)
        {
            $by -= 1;
            $bb += 1;
        }
        elsif($by == ($bb - 1))
        {
            $by -= 1;
        }
        
        my $bw = $br - $bx;
        my $bh = $bb - $by;
        
        if(!defined($bw) or ($bw <= 1))
        {
            $bw = 1;
        }
        if(!defined($bh) or ($bh <= 1))
        {
            $bh = 1;
        }
        
        my ($padLeft, $padTop, $padRight, $padBottom) = @{$self->{PADDING}};
        
        # if we are pure auto, size auto width/height to the perfect fit
        if($self->{AUTO_SCALE} == 1)
        {
            if($self->width eq 'auto')
            {
                $self->width($padLeft + $padRight + $bw);
            }
            if($self->height eq 'auto')
            {
                $self->height($padTop + $padBottom + $bh);
            }
        }
        
        # here, anything that remains auto needs to have some extra magic
        #  going on.  pretend the real width/height is a perfect fit.
        # ix = 0, iy = 0, ir = iw, ib = iy
        my ($rw, $aw);
        if($self->width eq 'auto')
        {
            $rw = $bw;
            $aw = -1;
        }
        else
        {
            $rw = $self->width - $padLeft - $padRight;
            $aw = $rw / $bw;
        }
        my ($rh, $ah);
        if($self->height eq 'auto')
        {
            $rh = $bh;
            $ah = -1;
        }
        else
        {
            $rh = $self->height - $padTop - $padBottom;
            $ah = $rh / $bh;
        }

        if($self->{AUTO_SCALE} == 2)
        {
            if($aw == -1)
            {
                $aw = $ah;
            }
            elsif($ah == -1)
            {
                $ah = $aw;
            }
            elsif($aw < $ah)
            {
                $ah = $aw;
            }
            else
            {
                $aw = $ah;
            }
        
            if($self->width eq 'auto')
            {
                $rw = $bw * $aw;
                $self->width($padLeft + $padRight + $rw);
            }
            elsif($self->height eq 'auto')
            {
                $rh = $bh * $ah;
                $self->height($padTop + $padBottom + $rh);
            }
            else # no-one was auto, let's modify padding then.
            {
                $rw = $bw * $aw;
                $rh = $bh * $ah;
                
                my $extraWidth  = $self->width - $padLeft - $padRight - $rw;
                my $extraHeight = $self->height - $padTop - $padBottom - $rh;
                
                
                if($padLeft + $padRight == 0)
                {
                    # center if we don't know what else to do
                    $padLeft = $extraWidth / 2;
                }
                else
                {
                    $padLeft += $extraWidth *
                                ($padLeft / ($padLeft + $padRight));
                }
                
                if($padTop + $padBottom == 0)
                {
                    # center if we don't know what else to do
                    $padTop = $extraHeight / 2;
                }
                else
                {
                    $padTop += $extraHeight *
                               ($padTop / ($padTop + $padBottom));
                }                
            }
        }
        
        push(@$lenses, sub
            {
                my ($x, $y) = @_;
                $x = (($x - $bx) / $bw * $rw + $padLeft);
                $y = (($y - $by) / $bh * $rh + $padTop);
                
                return ($x, $y);
            });
    }
    
    return $lenses;
}

sub draw
{
    my $self = shift;
    #my ($filename) = @_;
    
    my $engine =
        OnlyTheWind::KoalaRainbow::DrawEngines::EngineFactory->makeEngine();

    # bounds determines the width and height if set to auto.
    my $bounds = $self->calculateBounds();
    my $layers = $self->calculateLayers();

    my $lenses = $self->_makeLenses();
    
    $engine->init('width' => $self->{WIDTH},
                  'height' => $self->{HEIGHT},
                  'background' => $self->{BACKGROUND});

    $engine->lenses($lenses);
    
    my $useImageMap = defined($self->{IMAGE_MAP_NAME});
    my $imageMapEngine = undef;

    if($useImageMap)
    {
        $imageMapEngine =
            OnlyTheWind::KoalaRainbow::DrawEngines::ImageMap->new();
        $imageMapEngine->lenses($lenses);
    }

    my $filtersByLayer = $self->{FILTERS};

    for my $layer (sort {$a <=> $b} keys(%$layers))
    {
        my $layerNodes = $layers->{$layer};
        for my $node (@$layerNodes)
        {
            $node->draw($engine);
            if($useImageMap)
            {
                $node->drawImageMap($imageMapEngine);
            }
        }
        
        my $layerFilters = $filtersByLayer->{$layer};
        
        for my $filter (@$layerFilters)
        {
            $engine->filter($filter);
        }
    }
    
    if($useImageMap)
    {
        $self->imageMapContents($imageMapEngine->gimmeMap($self->imageMapName));
    }
    
    return $engine->renderToBlob();
}

1;
