#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::FuncRegistry;
use strict;

use OnlyTheWind::KoalaRainbow::DrawNodes::Color;
use OnlyTheWind::KoalaRainbow::ContextManager;

use OnlyTheWind::KoalaRainbow::Util qw( applyTimeStampDelta parseTimeInterval
        thresholdTimeStamp nowTimeStamp timeStampToSecs secsToTimeStamp
       );

use MT::Util qw(dirify format_ts);

use POSIX qw(floor);

my $funcRegistry = {};

# registerd functions have a number of attributes, they aren't just a reference.
# The supported attributes are:
# 'canPropagateConstraints' - If defined, the value should be a function
#   reference


sub registerFuncs
{
    # we may be called locally or through a package reference... get rid of
    #  the classname if it's there.  (assuming people use proper number of
    #  args.)
    if((@_ % 2) == 1)
    {
        shift;
    }
    my %funcs = @_;
    
    for my $funcName (keys %funcs)
    {
        $funcRegistry->{$funcName} = $funcs{$funcName};
    }
}

sub hasFunc
{
    my ($classname, $funcname) = @_;

    return exists $funcRegistry->{$funcname};
}

sub getFunc
{
    my ($classname, $funcname) = @_;
    my $funcInfo = $funcRegistry->{$funcname};
    
    return $funcInfo->{func};
}

sub getFuncAttrs
{
    my ($classname, $funcname) = @_;
    my $funcInfo = $funcRegistry->{$funcname};
    
    return $funcInfo;
}

sub getFuncAttr
{
    my ($classname, $funcname, $attrname) = @_;
    my $funcInfo = $funcRegistry->{$funcname};
    
    return $funcInfo->{$attrname};
}

registerFuncs(
              ##### math #####
              'int' => 
                  {isScalar => 1, argsNumbers => 1, func => \&castInt},
              'add' =>
                  {isScalar => 1, argsNumbers => 1, func => \&add},
              'sub' =>
                  {isScalar => 1, argsNumbers => 1, func => \&subtract},
              'mul' =>
                  {isScalar => 1, argsNumbers => 1, func => \&multiply},
              'div' =>
                  {isScalar => 1, argsNumbers => 1, func => \&divide},
              'min' =>
                  {isScalar => 1, argsNumbers => 1, func => \&doMin},
              'max' =>
                  {isScalar => 1, argsNumbers => 1, func => \&doMax},
              'clamp' =>
                  {isScalar => 1, argsNumbers => 1, func => \&clamp},
              'lerp' =>
                  {isScalar => 1, argsNumbers => 1, func => \&lerp},
              'num_range' =>
                  {isScalar => 0, argsNumbers => 1, func => \&numRange},
              ##### strings #####
              'len' =>
                  {argsStrings => 1, func => \&len},
              'concat' =>
                  {argsStrings => 1, func => \&concat},

              ##### MT hacks #####
              'mtdirify' =>
                  {argsStrings => 1, func => \&mtDirify},              
              ##### sets #####
              'count' =>
                  {needContext => 1, isScalar => 1, argsSets => 1, func => \&count},         
              'flatten' =>
                  {isScalar => 0, argsSets => 1, func => \&flattenContext},
              'val_in_set' =>
                  {isScalar => 1, argsAny => 1, func => \&val_in_set},
              ##### context #####
              'self' =>
                  {needContext => 1, func => \&self,
                   nondeterministic => 1},
              'parent' =>
                  {needContext => 1, func => \&parent,
                   nondeterministic => 1},                          
              'mtenv' =>
                  {needContext => 1, func => \&mtenv},
              'position' =>
                  {needContext => 1, argsAny => 1, func => \&position,
                   propConstraintWithComp => \&propPosition},
              'position_percent' =>
                  {needContext => 1, argsAny => 1, func => \&position_percent},                  
              'position_percent_zero' =>
                  {needContext => 1, argsAny => 1, 
                     func => \&position_percent_zero},
              ##### date #####
              'now' =>
                  {needContext => 1, func => \&now},
              'minutes_old' =>
                  {needContext => 1, func => \&minutesOld,
                   propConstraintWithComp => \&propMinutesOld},
              'hours_old' =>
                  {needContext => 1, func => \&hoursOld,
                   propConstraintWithComp => \&propHoursOld},
              'days_old' =>
                  {needContext => 1, func => \&daysOld,
                   propConstraintWithComp => \&propDaysOld},
              'date_thresh' =>
                  {func => \&dateThresh},
              'date_manip' =>
                  {func => \&dateManip},
              'date_in_range' =>
                  {func => \&dateInRange,
                   propConstraintsStandalone => \&propDateInRange},              
              'dates_to_range' => 
                  {func => \&datesToRange},        
              'date_range' =>
                  {func => \&dateRange},
              'date_range_set' =>
                  {func => \&dateRangeSet},
              'fmtdate' =>
                  {needContext => 1, func => \&fmtdate},
              ##### colors #####
              'color' =>
                  {func => \&color},
              'rgb' =>
                  {func => \&rgb},
              'hsv' =>
                  {func => \&hsv},
              'gray' =>
                  {func =>  \&gray},
              'grey' =>
                  {func => \&gray},
              );

sub fmtdate
{
    my ($ctxMgr, $format_string, $timestamp) = @_;
    
    return format_ts($format_string, $timestamp, $ctxMgr->findHint('blog'));
}

sub parent
{
    my $ctxMgr = shift;
    my ($up) = @_;
    
    $up = 1 if(!defined($up));
    
    for(;$up > 0; $up--)
    {
        if(defined($ctxMgr->parent))
        {
            $ctxMgr = $ctxMgr->parent;
        }
        elsif(defined($ctxMgr->owner))
        {
            $ctxMgr = $ctxMgr->owner;
        }
    }
    
    #warn 'num nodes in parent: ' . scalar(@{$ctxMgr->nodes});
    
    return $ctxMgr;
}

sub self
{
    my $ctxMgr = shift;
    my ($up) = @_;
    
    if(defined($up))
    {
        for(;$up > 0; $up--)
        {
            if(defined($ctxMgr->parent))
            {
                $ctxMgr = $ctxMgr->parent;
            }
            elsif(defined($ctxMgr->owner))
            {
                $ctxMgr = $ctxMgr->owner;
            }
        }
    }
    
    return $ctxMgr->nodes->[0];
}

sub val_in_set
{
    my ($val, $set) = @_;
    
    if(ref($val) eq 'OnlyTheWind::KoalaRainbow::ContextManager')
    {
        $val = $val->nodes->[0];
    }
    
    for my $cval (@{$set->nodes})
    {
        if($val == $cval)
        {
            return 1;
        }
    }
    
    return 0;
}

sub count
{
    # so, our first arg is provided due to needsContext, and may not be the
    #  one the count is for.
    my $ctxMgr = shift;
    
    my $set;
    # if there is another arg, that's the thing to count.
    if(@_)
    {
        $set = shift;
    }
    else # if not, use our current context.
    {
        $set = $ctxMgr;
    }

    if(ref($set) eq 'OnlyTheWind::KoalaRainbow::ContextManager')
    {
        $set = $set->nodes;
    }
    
    if(defined($set))
    {
        return scalar(@{$set});
    }
    else
    {
        warn 'Tried to take count of undef variable. Returning 0.';
        return 0;
    }
}

sub flatten
{
    my $ctxMgr = shift;
    
    my $outNodes = [];
    
    for my $node (@{$ctxMgr->nodes})
    {
        if(ref($node) eq 'ARRAY')
        {
            push(@$outNodes, @$node);
        }
        else
        {
            push(@$outNodes, $node);
        }
    }
    
    my $newCtxMgr = OnlyTheWind::KoalaRainbow::ContextManage->new($outNodes);
    $newCtxMgr->parent($ctxMgr);
    
    return $newCtxMgr;
}

sub now
{
    my $ctxMgr = shift;
    
    # we need to tell it the blog
    my $blog = $ctxMgr->findHint('mtcontext')->stash('blog');
    
    return nowTimeStamp($blog);
}

sub propTimeOld
{
    my ($ctxMgr, $constraints, $argNodes, $expNode, $secsInInterval) = @_;

    my $blog = $ctxMgr->findHint('mtcontext')->stash('blog');
    
    if((scalar(@$argNodes) == 1) &&
       $constraints->isEdgeNode($argNodes->[0]))
    {
        # --- get now, get the number we're being compared against, multiply
        #  that the number of seconds in the interval, then add that to the
        #  seconds to get the timestamp, tweaking by one second if the
        #  comparison is equalish (direction depending on the comparison).
        #  (We need to do this because the 
        #  Defer to the constraint object to figure out how to handle the actual
        #  range bounds.
        my $secs = time;
        
        # we are guaranteed the constant is on the right side of expNode, 
        #  and that is is a constant.
        my $deltaSecs = $expNode->right->evaluateValue($ctxMgr) *
                        $secsInInterval;

        my $threshVal = $secs - $deltaSecs;
        
        my $threshTimestamp = secsToTimeStamp($blog, $threshVal);
        
        my $op = $expNode->op;
        
        # since the constraints are non-inclusive, we need to bump by a second
        #  to meet our inclusive needs if it's a than-equalish comparison.
        if($expNode->isThanEqualishOp())
        {
            if($op eq '<=')
            {
                $threshVal += 1;
                $op = '<';
            }
            elsif($op eq '>=')
            {
                $threshVal -= 1;
                $op = '>';
            }
        }
    
        if($constraints->addComparisonConstraint($argNodes->[0],
                                                 $op, $threshTimestamp))
        {
            $expNode->mark(1);
            return 1;
        }
    }
    
    return 0;
}

sub minutesOld
{
    my ($ctxMgr, $timestamp) = @_;
    
    # we need to tell it the blog
    my $blog = $ctxMgr->findHint('mtcontext')->stash('blog');
    
    my $deltaSecs = time - timeStampToSecs($blog, $timestamp);
    
    return int($deltaSecs / 60);
}

sub propMinutesOld
{
    return propTimeOld(@_, 60);
}

sub hoursOld
{
    my ($ctxMgr, $timestamp) = @_;
    
    # we need to tell it the blog
    my $blog = $ctxMgr->findHint('mtcontext')->stash('blog');
    
    my $deltaSecs = time - timeStampToSecs($blog, $timestamp);
    
    return int($deltaSecs / (60 * 60));
}

sub propHoursOld
{
    return propTimeOld(@_, 60 * 60);
}


sub daysOld
{
    my ($ctxMgr, $timestamp) = @_;

    # we need to tell it the blog
    my $blog = $ctxMgr->findHint('mtcontext')->stash('blog');
    
    my $deltaSecs = time - timeStampToSecs($blog, $timestamp);
    
    return int($deltaSecs / (60 * 60 * 24));
}

sub propDaysOld
{
    return propTimeOld(@_, 60 * 60 * 24);
}


sub dateThresh
{
    return thresholdTimeStamp(@_);
}

sub dateManip
{
    return applyTimeStampDelta(@_);
}


sub datesToRange
{
    my ($date1, $date2) = @_;
    return [$date1, $date2];
}

sub dateInRange
{
    my ($timestamp, $ts_range) = @_;
    
    # we expect the first one to before the second one.  If they flip-flop,
    #  then this will handle it.  If they are equal, we are screwed by this
    #  logic, so become -1 in such a case (either is fine.)
    my $less_than = ($ts_range->[0] cmp $ts_range->[1]) || -1;
    
    my $retval = 0;
    if((($timestamp cmp $ts_range->[0]) != $less_than) and
       (($timestamp cmp $ts_range->[1]) != -$less_than))
    {
        $retval = 1;
    }    
    
    return $retval;
}

sub dateRange
{
    my ($timestamp, $threshold, $delta, $tweakDelta) = @_;
    if(!defined($tweakDelta))
    {
        # pay attention to the most significant sign, and do the opposite.
        if(substr($delta, 0, 1) eq '-')
        {
            $tweakDelta = '1s';
        }
        else
        {
            $tweakDelta = '-1s';
        }
    }
    
    my $startTime = thresholdTimeStamp($timestamp, $threshold);
    my $endTime = applyTimeStampDelta($startTime, $delta);
    # now tweak the end time.
    $endTime = applyTimeStampDelta($endTime, $tweakDelta);
    return [$startTime, $endTime];
}

sub dateRangeSet
{
    my ($timestamp, $threshold, $delta, $numInSet, $setDelta, $tweakDelta) = @_;

    if(!defined($tweakDelta))
    {
        # pay attention to the most significant sign, and do the opposite.
        if(substr($delta, 0, 1) eq '-')
        {
            $tweakDelta = '1s';
        }
        else
        {
            $tweakDelta = '-1s';
        }
    }
    
    my $ranges = [];
    
    my $startTime = thresholdTimeStamp($timestamp, $threshold);
    
    for(;$numInSet > 0; $numInSet--)
    {
        my $endTime = applyTimeStampDelta($startTime, $delta);
        $endTime = applyTimeStampDelta($endTime, $tweakDelta);

        push(@$ranges, [$startTime, $endTime]);

        $startTime = applyTimeStampDelta($startTime, $setDelta);
    }
    
    return OnlyTheWind::KoalaRainbow::ContextManager->new($ranges);
}

sub castInt
{
    # this is just a trick to make the function handler coerce the type for us.
    # of course, something is required to do this.  The problem is that some
    #  values like comment_count may not actually be '0' when they are zero;
    #  this mechanism allows us to force it into an integer without requiring
    #  the receiving code to be all super-smart.  There are probably few cases
    #  where the use of this function is not representative of something that
    #  should be fixed.  If this is required, look for something that should
    #  probably coerce as needed...
    
    # okay, now I went and documented that we turn empty strings into zeroes, so
    #  we need to do that too.
    my $allegedNum = shift;
    if($allegedNum eq '')
    {
        return 0;
    }
    else
    {
        return $allegedNum;
    }
}

sub doMin
{
    my $answer = shift;
    while(@_)
    {
        my $newval = shift;
        
        if($newval < $answer)
        {
            $answer = $newval;
        }
    }
    
    return $answer;
}

sub doMax
{
    my $answer = shift;
    while(@_)
    {
        my $newval = shift;
        
        if($newval > $answer)
        {
            $answer = $newval;
        }
    }
    
    return $answer;
}

sub add
{
    my $answer = shift;
    while(@_)
    {
        $answer += shift;
    }
    
    return $answer;
}

sub subtract
{
    my $answer = shift;
    while(@_)
    {
        $answer -= shift;
    }
    
    return $answer;
}

sub multiply
{
    my $answer = shift;
    while(@_)
    {
        $answer = $answer * shift;
    }
    
    return $answer;
}

sub divide
{
    my ($left, $right) = @_;
    return $left / $right;
}

sub clamp
{
    my ($uval, $umin, $umax) = @_;
        
    # clamp uval
    if($uval < $umin)
    {
        $uval = $umin;
    }
    if($uval > $umax)
    {
        $uval = $umax;
    }
    
    return $uval;
}

sub lerp
{
    my ($uval, $umin, $umax, $omin, $omax) = @_;
        
    my $urange = $umax - $umin;
    if($urange == 0)
    {
        $urange = 1;
    }
    my $orange = $omax - $omin;
    
    # clamp uval
    if($uval < $umin)
    {
        $uval = $umin;
    }
    if($uval > $umax)
    {
        $uval = $umax;
    }
    
    return (((($uval - $umin) / $urange) * $orange) + $omin);
}

sub len
{
    my $text = shift;
    
    return length($text);
}

sub concat
{
    my $answer = shift;
    # actually, I think I could probably just use join here, but I expect
    #  per value processing will eventually be required, so here we are.
    while(@_)
    {
        $answer .= shift;
    }
    
    return $answer;
}

sub mtDirify
{
    return dirify(@_);
}

sub mtenv
{
    my $ctxMgr = shift;
    my $key = shift;
    
    my $putInContext = 1;
    if(@_)
    {
        $putInContext = shift;
    }
    
    my $val = $ctxMgr->findHint('mtcontext')->stash($key);
    
    # convert arrays into ContextManagers
    if(ref($val) eq 'ARRAY')
    {
        my $newCtxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new($val);
        $newCtxMgr->parent($ctxMgr);
        
        $val = $newCtxMgr;
    }
    elsif($putInContext)
    {
        my $newCtxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new([$val]);
        $newCtxMgr->parent($ctxMgr);
        
        $val = $newCtxMgr;        
    }
    
    return $val;
}

sub _positionContextHelper
{
    my ($objCtxMgr, $ctxMgr) = @_;
    my $obj;
    
    if(ref($objCtxMgr) eq 'OnlyTheWind::KoalaRainbow::ContextManager')
    {
        $obj = $objCtxMgr->nodes->[0];
    }
    else
    {
        $obj = $objCtxMgr;
    }

    if(!defined($obj) or !defined($ctxMgr))
    {
        return (0, 0);
    }
    
    my $useId;
    my $cmpItem;
    if(substr(ref($obj), 0, 4) eq 'MT::')
    {
        $useId = 1;
        $cmpItem = $obj->id;
    }
    else
    {
        $useId = 0;
        $cmpItem = $obj;
    }
    
    my $count = scalar(@{$ctxMgr->nodes});
    my $index = 0;
    for my $node (@{$ctxMgr->nodes})
    {

        my $nodeItem;        
        if($useId)
        {        
            $nodeItem = $node->id;
        }
        else
        {
            $nodeItem = $node;
        }
        
        if(($nodeItem == $cmpItem) or ($nodeItem eq $cmpItem))
        {
            return ($index, $count);
        }
        
        $index++;
    }

    return (0, $count); # we will document our lack of support for looking for
       # things that don't exist.
}

sub position
{
    my $ctxMgr = shift;

    if(@_)
    {
        my ($index, $count) = _positionContextHelper(@_);
        return $index;
    }
    else
    {
        return $ctxMgr->hint('position');
    }
}

sub position_percent
{
    my $ctxMgr = shift;
    
    my $pos;
    my $count;
    
    if(@_)
    {
        ($pos, $count) = _positionContextHelper(@_);
        # our dynamics mandate this.
        $pos++;
    }
    else
    {
        $pos = $ctxMgr->hint('position') + 1;
        $count = $ctxMgr->hint('count');
    }
    
    if($count == 0)
    {
        return 0.0;
    }
    else
    {
        return $pos / $count;
    }
}

sub position_percent_zero
{
    my $ctxMgr = shift;
    
    my $pos;
    my $count;
    
    if(@_)
    {
        ($pos, $count) = _positionContextHelper(@_);
        # our dynamics mandate this.
    }
    else
    {
        $pos = $ctxMgr->hint('position');
        $count = $ctxMgr->hint('count');
    }
    
    if($count == 0)
    {
        return 0.0;
    }
    else
    {
        return $pos / $count;
    }
}


sub color
{
    my $colorName = shift;
    
    my $color = OnlyTheWind::KoalaRainbow::DrawNodes::Color->new();
    
    if(substr($colorName, 0, 1) eq '#')
    {
        $color->hexcode($colorName);
    }
    else
    {
        $color->name($colorName);
    }
    
    return $color;
}

sub rgb
{
    my ($r, $g, $b) = @_;
    
    my $color = OnlyTheWind::KoalaRainbow::DrawNodes::Color->new();
    
    $color->rgb($r, $g, $b);
    
    return $color;
}

# adapted from my goo samurui hsv routine, pardon any absurdities.
# all values should be floats be in the range [0.0, 1.0]
sub hsv
{
    my ($hue, $saturation, $brightness) = @_;

    my $h = ($hue - floor($hue)) * 6.0;
    my $fh = floor($h);
    my $f = $h - $fh;
    my $p = $brightness * (1.0 - $saturation);
    my $q = $brightness * (1.0 - ($saturation * $f));
    my $t = $brightness * (1.0 - ($saturation * (1.0 - $f)));

    if($fh == 0)
    {
        return rgb(floor(($brightness * 255.0) + 0.5),
                   floor(($t * 255.0) + 0.5),
                   floor(($p * 255.0) + 0.5));
    }
    elsif($fh == 1)
    {
        return rgb(floor(($q * 255.0) + 0.5),
                   floor(($brightness * 255.0) + 0.5),
                   floor(($p * 255.0) + 0.5));
    }
    elsif($fh == 2)
    {
        return rgb(floor(($p * 255.0) + 0.5),
                   floor(($brightness * 255.0) + 0.5),
                   floor(($t * 255.0) + 0.5));
    }
    elsif($fh == 3)
    {
        return rgb(floor(($p * 255.0) + 0.5),
                   floor(($q * 255.0) + 0.5),
                   floor(($brightness * 255.0) + 0.5));
    }
    elsif($fh == 4)
    {
        return rgb(floor(($t * 255.0) + 0.5),
                   floor(($p * 255.0) + 0.5),
                   floor(($brightness * 255.0) + 0.5));
    }
    elsif($fh == 5)
    {
        return rgb(floor(($brightness * 255.0) + 0.5),
                   floor(($p * 255.0) + 0.5),
                   floor(($q * 255.0) + 0.5));
    }                
    # a technical impossibility to be here, I'm sure.
}

sub gray
{
    my $shade = shift;
    
    my $color = OnlyTheWind::KoalaRainbow::DrawNodes::Color->new();
    
    $color->rgb($shade, $shade, $shade);
    
    return $color;  
}


1;
