#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::GlobalSetRegistry;
use strict;

use OnlyTheWind::KoalaRainbow::ContextManager;

use MT::Entry;
use MT::Blog;
use MT::Author;
use MT::Category;
use MT::TBPing;
use MT::IPBanList;
use MT::Comment;
use MT::Placement;
use MT::Notification;

#use Data::Dump qw(dump);

my $globalSetRegistry = {};

sub registerGlobalSets
{
    # we may be called locally or through a package reference... get rid of
    #  the classname if it's there.  (assuming people use proper number of
    #  args.)
    if((@_ % 2) == 1)
    {
        shift;
    }
    my %sets = @_;
    
    for my $setName (keys %sets)
    {
        $globalSetRegistry->{$setName} = $sets{$setName};
    }
}

registerGlobalSets('blogs' =>
                    {'factory' => \&loadBlogs,
                     'legalEdges' => MT::Blog->properties->{columns}},
                   'categories' =>
                    {'factory' => \&loadCategories,
                     'legalEdges' => MT::Category->properties->{columns}},
                   'entries' =>
                    {'factory' => \&loadEntries,
                     'legalEdges' => MT::Entry->properties->{columns}},
                   'authors' =>
                    {'factory' => \&loadAuthors,
                     'legalEdges' => MT::Author->properties->{columns}},
                   'trackbacks' =>
                    {'factory' => \&loadTrackbacks,
                     'legalEdges' => MT::TBPing->properties->{columns}},
                   'bannedips' =>
                    {'factory' => \&loadBannedIPs,
                     'legalEdges' => MT::IPBanList->properties->{columns}},
                   'comments' =>
                    {'factory' => \&loadComments,
                     'legalEdges' => MT::Comment->properties->{columns}},
                   'placements' =>
                    {'factory' => \&loadPlacements,
                     'legalEdges' => MT::Placement->properties->{columns}},
                   'notifications' =>
                    {'factory' => \&loadNotifications,
                     'legalEdges' => MT::Notification->properties->{columns}},
                  );

sub getSetLoader
{
    my ($class, $setname) = @_;
    
    #die 'Returning set loader for "' . $setname . '"';
    #die dump($globalSetRegistry);
    return $globalSetRegistry->{$setname} -> {factory};
}

sub getSetInfo
{
    my ($class, $setname) = @_;
    
    return $globalSetRegistry->{$setname};
}

sub isAllowedEdge
{
    my ($class, $setname, $edgename) = @_;
    
    my $legalEdges = $globalSetRegistry->{$setname}->{legalEdges};
    
    return scalar(grep {$_ eq $edgename} @$legalEdges);
}

sub loadBlogs
{
    my ($constraintMgr) = @_;
    
    my $terms = {};
    my $arguments = {};
    
    if(defined($constraintMgr))
    {
        $terms     = $constraintMgr->terms;
        $arguments = $constraintMgr->arguments;
    }
    
    my @objects = MT::Blog->load($terms, $arguments);
    
    my $ctxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new(\@objects);
    
    return $ctxMgr;
}

sub loadCategories
{
    my @objects = MT::Category->load({}, {});
    
    my $ctxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new(\@objects);
    
    return $ctxMgr;
}

sub loadEntries
{
    my ($constraintMgr) = @_;
    
    my $terms = {};
    my $arguments = {};
    
    if(defined($constraintMgr))
    {
        $terms     = $constraintMgr->terms;
        $arguments = $constraintMgr->arguments;
    }

    my @objects = MT::Entry->load($terms, $arguments);
    
    my $ctxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new(\@objects);
    
    return $ctxMgr;
}

sub loadAuthors
{
    my ($constraintMgr) = @_;
    
    my $terms = {};
    my $arguments = {};
    
    if(defined($constraintMgr))
    {
        $terms     = $constraintMgr->terms;
        $arguments = $constraintMgr->arguments;
    }

    my @objects = MT::Author->load($terms, $arguments);
    
    my $ctxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new(\@objects);
    
    return $ctxMgr;
}

sub loadTrackbacks
{
    my ($constraintMgr) = @_;
    
    my $terms = {};
    my $arguments = {};
    
    if(defined($constraintMgr))
    {
        $terms     = $constraintMgr->terms;
        $arguments = $constraintMgr->arguments;
    }

    my @objects = MT::TBPing->load($terms, $arguments);
    
    my $ctxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new(\@objects);
    
    return $ctxMgr;
}

sub loadBannedIPs
{
    my ($constraintMgr) = @_;
    
    my $terms = {};
    my $arguments = {};
    
    if(defined($constraintMgr))
    {
        $terms     = $constraintMgr->terms;
        $arguments = $constraintMgr->arguments;
    }

    my @objects = MT::IPBanList->load($terms, $arguments);
    
    my $ctxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new(\@objects);
    
    return $ctxMgr;
}

sub loadComments
{
    my ($constraintMgr) = @_;
    
    my $terms = {};
    my $arguments = {};
    
    if(defined($constraintMgr))
    {
        $terms     = $constraintMgr->terms;
        $arguments = $constraintMgr->arguments;
    }

    my @objects = MT::Comment->load($terms, $arguments);
    
    my $ctxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new(\@objects);
    
    return $ctxMgr;
}

sub loadPlacements
{
    my ($constraintMgr) = @_;
    
    my $terms = {};
    my $arguments = {};
    
    if(defined($constraintMgr))
    {
        $terms     = $constraintMgr->terms;
        $arguments = $constraintMgr->arguments;
    }

    my @objects = MT::Placement->load($terms, $arguments);
    
    my $ctxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new(\@objects);
    
    return $ctxMgr;
}

sub loadNotifications
{
    my ($constraintMgr) = @_;
    
    my $terms = {};
    my $arguments = {};
    
    if(defined($constraintMgr))
    {
        $terms     = $constraintMgr->terms;
        $arguments = $constraintMgr->arguments;
    }

    my @objects = MT::Notification->load($terms, $arguments);
    
    my $ctxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new(\@objects);
    
    return $ctxMgr;
}

1;