#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::QueryParseNodes::Expression;
use strict;

# Note that at this time, expressions are only capable of performing comparison
#  and logical operations.

sub new
{
    my $class = shift;
    
    my $self  =
    {
        LEFT  => undef,
        OP    => undef,
        RIGHT => undef,
        MARK  => undef,
    };
    
    bless ($self, $class);
    return $self;
}

sub clone
{
    my $self = shift;
    
    my $newMe = OnlyTheWind::KoalaRainbow::QueryParseNodes::Expression->new();
    $newMe->left($self->left()->clone());
    # op isn't actually a node, so just spit it out verbatim
    $newMe->op($self->op());
    
    if(defined($self->right()))
    {
        $newMe->right($self->right()->clone());
    }
    
    return $newMe;
}

sub left
{
    my $self = shift;
    
    if (@_) 
    {
        $self->{LEFT} = shift;
    }
    
    return $self->{LEFT};
}

sub right
{
    my $self = shift;
    
    if (@_) 
    {
        $self->{RIGHT} = shift;
    }
    
    return $self->{RIGHT};
}

sub op
{
    my $self = shift;
    
    if (@_) 
    {
        $self->{OP} = shift;
    }
    
    return $self->{OP};
}

sub mark
{
    my $self = shift;
    
    if (@_) 
    {
        $self->{MARK} = shift;
    }
    
    return $self->{MARK};
}

sub toString
{
    my $self = shift;

    my $str = '';
    
    if($self->{MARK})
    {
        $str .= '<<';
    }

    if($self->{OP})
    {
        $str .= '(' . $self->{LEFT}->toString() . ' ' . $self->{OP} . ' ' .
                $self->{RIGHT}->toString() . ')';
    }
    else
    {
        $str .= $self->{LEFT}->toString();
    }
    
    if($self->{MARK})
    {
        $str .= '>>';
    }
    return $str;
}

sub isLogicalOp
{
    my $self = shift;
    
    my $op = $self->op;
    return ($op eq 'and') or ($op eq 'or');
}

sub isComparisonOp
{
    my $self = shift;
    
    my $op = $self->op;
    return ($op eq '=') or ($op eq '!=') or
           ($op eq '<') or ($op eq '<=') or
           ($op eq '>') or ($op eq '>=');
}

sub isLessishOp
{
    my $self = shift;
    
    my $op = $self->op;
    return ($op eq '<') or ($op eq '<=');
}

sub isGreaterishOp
{
    my $self = shift;
    
    my $op = $self->op;
    return ($op eq '>') or ($op eq '>=');
}

sub isThanEqualishOp
{
    my $self = shift;
    
    my $op = $self->op;
    return ($op eq '<=') or ($op eq '>=');
}

sub isEqualishOp
{
    my $self = shift;
    
    my $op = $self->op;
    return ($op eq '=') or ($op eq '!=') or
           ($op eq '<=') or ($op eq '>=');
}

sub isConstant
{
    my $self = shift;
    
    if($self->op)
    {
        return $self->left->isConstant() and $self->right->isConstant();
    }
    else
    {
        return $self->left->isConstant();
    }
}

sub isScalar
{
    return 1;
}

# First, we can only propagate constraints from always-on AND clauses.
# Next, there are basically two cases that we need to deal with:
# - Direct comparisons between edges and constants (which may be numeric
#   constants, string literals, variables, or functions that evaluate to a
#   constant (which hinges on an appropriate annotation and all arguments
#   being constants))
# - Calls to functions that:
#  - Return booleans and can propagate in a standalone fashion.
#  - Return values but know how to consume a comparison and propagate the
#    constraints in that fashion.
sub propagateConstraints
{
    my $self = shift;
    my ($ctxMgr, $constraints) = @_;
    
    my $op = $self->op;

    # just pass the call along if we have no op
    if(!defined($op))
    {
        return $self->left->propagateConstraints($ctxMgr, $constraints);
    }
    
    # If we hit an 'or', we can't try and propagate anyone below us.
    # If we hit a '!=', we can't do anything about it, due to a lack of support
    #  for such a constraint.
    if(($op eq 'or') or ($op eq '!='))
    {
        return 0;
    }
    
    # If this is an and, try and propagate both constraints
    if($op eq 'and')
    {
        my $propLeft  = $self->left->propagateConstraints($ctxMgr,
                                                          $constraints);
        my $propRight = $self->right->propagateConstraints($ctxMgr,
                                                           $constraints);
        
        if($propLeft and $propRight)
        {
            $self->mark(1);
        }
        
        return ($propLeft || $propRight);
    }
    
    # normalize now
    $self->normalize();
    # (op may be modified)
    $op = $self->op;
    
    # Right must be a constant now
    my $rightNode = $self->right;
    if(!$rightNode->isConstant())
    {
        return 0;
    }
    
    my $leftNode = $self->left;

    # We can handle edges
    if($constraints->isEdgeNode($leftNode))
    {
        if($constraints->addComparisonConstraint($leftNode, $op,
           $rightNode->evaluateValue($ctxMgr)))
        {
            $self->mark(1);
            return 1;
        }   
    }
    # we can handle propagatable functions
    elsif($constraints->isFuncNode($leftNode))
    {
        my $funcNode = $leftNode;
        if($constraints->isQueryContextNode($funcNode))
        {
            $funcNode = $funcNode->evalNode;
        }
        return $funcNode->propagateCompConstraints($ctxMgr, $constraints,
                                                   $self);
    }
    
    # we can't handle anything else though.
    return 0;
}

sub evaluate
{
    my ($self, $ctxMgr) = @_;

#print "Evaluating " . $self->toString() . "\n";
    
    # if we are marked, it means either our constraints were propagated, or
    #  we did partial evaluation and determined a value.  either way, return
    #  said value.
    if($self->{MARK})
    {
#print "MARKED BAILING\n";
        return $self->{MARK};
    }
    
    my $leftNode = $self->{LEFT};
    if(!defined($leftNode))
    {
        return undef;
    }
    
    my $leftVal = $leftNode->evaluateValue($ctxMgr);
    
    my $op = $self->{OP};
    if(!defined($op))
    {
        return $leftVal;
    }
    
    # handle logical cases short-circuiting
    if(($op eq 'and') and !$leftVal)
    {
        return 0;
    }
    elsif(($op eq 'or') and $leftVal)
    {
        return $leftVal;
    }
    
    my $rightNode = $self->{RIGHT};
    if(!defined($rightNode))
    {
        # we have no unary ops.
        return undef;
    }
    
    my $rightVal = $rightNode->evaluateValue($ctxMgr);
        
    if(($op eq 'and') or ($op eq 'or'))
    {
        # we already checked leftVal in the short-circuit mode
        # same case for 'and' and 'or' because of this
        return $rightVal;
    }
    
    if(!defined($rightVal) or !defined($leftVal))
    {
        return undef;
    }
    
    # okay, so an eval would make this all happy, and there are presumably
    #  more efficient ways than this, but hey. next revision.
    if($op eq '=')
    {
        # see if the values are actually numbers via regex
        if(($leftVal =~ m/^-?\d+(.\d*)?$/) and
           ($rightVal =~ m/^-?\d+(.\d*)?$/))
        {
            return $leftVal == $rightVal;
        }
        else
        {
            return $leftVal eq $rightVal;
        }
    }
    elsif($op eq '!=')
    {
        # see if the values are actually numbers via regex
        if(($leftVal =~ m/^-?\d+(.\d*)?$/) and
           ($rightVal =~ m/^-?\d+(.\d*)?$/))
        {
            return $leftVal != $rightVal;
        }
        else
        {
            return $leftVal ne $rightVal;
        }
    }
    elsif($op eq '<')
    {
        # see if the values are actually numbers via regex
        if(($leftVal =~ m/^-?\d+(.\d*)?$/) and
           ($rightVal =~ m/^-?\d+(.\d*)?$/))
        {
            return $leftVal < $rightVal;
        }
        else
        {
            return $leftVal lt $rightVal;
        }
    }
    elsif($op eq '>')
    {
        # see if the values are actually numbers via regex
        if(($leftVal =~ m/^-?\d+(.\d*)?$/) and
           ($rightVal =~ m/^-?\d+(.\d*)?$/))
        {
            return $leftVal > $rightVal;
        }
        else
        {
            return $leftVal gt $rightVal;
        }
    }
    elsif($op eq '>=')
    {
        # see if the values are actually numbers via regex
        if(($leftVal =~ m/^-?\d+(.\d*)?$/) and
           ($rightVal =~ m/^-?\d+(.\d*)?$/))
        {
            return $leftVal >= $rightVal;
        }
        else
        {
            return $leftVal ge $rightVal;
        }
    }
    elsif($op eq '<=')
    {
        # see if the values are actually numbers via regex
        if(($leftVal =~ m/^-?\d+(.\d*)?$/) and
           ($rightVal =~ m/^-?\d+(.\d*)?$/))
        {
            return $leftVal <= $rightVal;
        }
        else
        {
            return $leftVal le $rightVal;
        }    
    }
    else
    {
        # error
        return undef;
    }
}

sub evaluateValue
{
    my ($self, $ctxMgr) = @_;
    
    return $self->evaluate($ctxMgr);
}

sub evaluateSet
{
    my ($self, $ctxMgr) = @_;

    die "Can't evaluate expression " . $self->toString() . " as a set.";
}

# partialEvaluate is always implicitly dealing with values...
sub partialEvaluate
{
    my $self = shift;
    my ($ctxMgr) = @_;

    if(defined($self->left))
    {
        $self->left($self->left->partialEvaluate($ctxMgr));
    }
    if(defined($self->right))
    {
        $self->right($self->right->partialEvaluate($ctxMgr));
    }
    
    if($self->isConstant())
    {
        my $value = $self->evaluate($ctxMgr);
        $self->mark($value);
    }
    
    return $self;
}

# if we have a function or edge on one side and a constant on the other, then
#  normalize ourselves to ensure that the function/edge is on the left and the
#  constant is on the right.
sub normalize
{
    my $self = shift;
    
    if(defined($self->op))
    {
        if(((ref($self->right) eq
             'OnlyTheWind::KoalaRainbow::QueryParseNodes::FunctionCall') ||
            (ref($self->right) eq
             'OnlyTheWind::KoalaRainbow::QueryParseNodes::ContextVariable')) &&
           $self->left->isConstant())
        {
            # func on right, constant on left.  swap!
            my $left = $self->left();
            my $right = $self->right();
            $self->left($right);
            $self->right($left);
            
            # flip the op
            my $op = $self->op;
            # try crazy syntax some more
            $self->op('>')  if $op eq '<';
            $self->op('>=') if $op eq '<=';
            $self->op('<')  if $op eq '>';
            $self->op('<=') if $op eq '>=';
        }
    }
}

1;
