#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::QueryParseNodes::FunctionCall;
use strict;

use OnlyTheWind::KoalaRainbow::FuncRegistry;
use OnlyTheWind::KoalaRainbow::QueryParseNodes::DerivedValue;

sub new
{
    my $class = shift;
    
    my $self  =
    {
        NAME  => undef,
        ARGS  => undef,
    };
    
    bless ($self, $class);
    return $self;
}

# we need to clone the args
sub clone
{
    my $self = shift;
    
    my $newMe = OnlyTheWind::KoalaRainbow::QueryParseNodes::FunctionCall->new();
    $newMe->name($self->name());
    
    if(defined($self->args()))
    {
        my $newArgs = [];
        for my $arg (@{$self->args()})
        {
            push(@$newArgs, $arg->clone());
        }
        
        $newMe->args($newArgs);
    }
    
    return $newMe;
}

sub name
{
    my $self = shift;
    
    if (@_) 
    {
        $self->{NAME} = shift;
    }
    
    return $self->{NAME};
}

sub args
{
    my $self = shift;
    
    if (@_)
    {
        $self->{ARGS} = shift;
    }
    
    return $self->{ARGS};
}

sub toString
{
    my $self = shift;

    if($self->{NAME})
    {
        my $str = $self->{NAME} . '(';
        
        for my $arg (@{$self->{ARGS}})
        {
            $str .= $arg->toString();
        }
        continue
        {
            $str .= ', ';
        }
        
        $str .= ')';
        
        return $str;
    }
    else
    {
        return 'unhappy function call';
    }
}

sub partialEvaluate
{
    my ($self, $ctxMgr) = @_;

    my $funcName = $self->{NAME};
    my $argNodes = $self->{ARGS};

    # call partialEvaluate on all arguments
    my $newArgNodes = [];
    for my $argNode (@$argNodes)
    {
        push(@$newArgNodes, $argNode->partialEvaluate($ctxMgr));
    }
    $argNodes = $newArgNodes;
    $self->{ARGS} = $argNodes;

    # bail if not all of the arguments are constants
    for my $argNode (@$argNodes)
    {
        if(!$argNode->isConstant())
        {
            return $self;
        }
    }

    # get the function attributes
    my $funcAttrs = 
           OnlyTheWind::KoalaRainbow::FuncRegistry->getFuncAttrs($funcName);

    # bail if it isn't deterministic
    if($funcAttrs->{nondeterministic})
    {
        return $self;
    }

    my $val = $self->evaluate($ctxMgr);
    return OnlyTheWind::KoalaRainbow::QueryParseNodes::DerivedValue->new($val);
}

# caller is responsible for only calling in a boolean context
sub propagateConstraints
{
    my ($self, $ctxMgr, $constraints) = @_;

    my $funcName = $self->{NAME};
    my $argNodes = $self->{ARGS};

    my $funcAttrs = 
           OnlyTheWind::KoalaRainbow::FuncRegistry->getFuncAttrs($funcName);
           
    if(defined($funcAttrs->{propConstraintsStandalone}))
    {
        my $propAloneFunc = $funcAttrs->{propConstraintsStandalone};
        
        return $propAloneFunc->($ctxMgr, $constraints, $argNodes);
    }
    
    return 0;
}

sub propagateCompConstraints
{
    my ($self, $ctxMgr, $constraints, $expNode) = @_;

    my $funcName = $self->{NAME};
    my $argNodes = $self->{ARGS};

    my $funcAttrs = 
           OnlyTheWind::KoalaRainbow::FuncRegistry->getFuncAttrs($funcName);

    if(defined($funcAttrs->{propConstraintWithComp}))
    {
        my $propCompFunc = $funcAttrs->{propConstraintWithComp};
        
        return $propCompFunc->($ctxMgr, $constraints, $argNodes, $expNode);
    }
    
    return 0;
}

sub evaluate
{
    my ($self, $ctxMgr) = @_;

    my $funcName = $self->{NAME};
    my $argNodes = $self->{ARGS};
    
    if(!OnlyTheWind::KoalaRainbow::FuncRegistry->hasFunc($funcName))
    {
        die 'No such function as ' . $funcName;
    }
    
    my $funcAttrs = 
           OnlyTheWind::KoalaRainbow::FuncRegistry->getFuncAttrs($funcName);
    
    my @argVals = ();
    
    if($funcAttrs->{needContext})
    {
        push(@argVals, $ctxMgr);
    }
    
    # get argument values
    for my $argNode (@$argNodes)
    {
        my $argVal;
        
        if(defined($funcAttrs->{argsSets}))
        {
            $argVal = $argNode->evaluateSet($ctxMgr);
        }
        elsif(defined($funcAttrs->{argsAny}))
        {
            $argVal = $argNode->evaluate($ctxMgr);
        }
        else
        {
            $argVal = $argNode->evaluateValue($ctxMgr);
        }
        
        if($funcAttrs->{argsNumbers})
        {
            $argVal = 0 if !defined($argVal);            
        }
        elsif($funcAttrs->{argsStrings})
        {
            $argVal = '' if !defined($argVal);
        }

if(!defined($argVal))
{
    warn 'undefined node was: ' . $argNode->toString();
}
        
        push(@argVals, $argVal);
    }

    return $funcAttrs->{func}->(@argVals);
}

sub evaluateValue
{
    my ($self, $ctxMgr) = @_;

    return $self->evaluate($ctxMgr);
}

sub evaluateSet
{
    my ($self, $ctxMgr) = @_;

    return $self->evaluate($ctxMgr);
}

1;
