#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::QueryParser;
use strict;

use OnlyTheWind::KoalaRainbow::QueryParseNodes::ContextVariable;
use OnlyTheWind::KoalaRainbow::QueryParseNodes::Expression;
use OnlyTheWind::KoalaRainbow::QueryParseNodes::FunctionCall;
use OnlyTheWind::KoalaRainbow::QueryParseNodes::Null;
use OnlyTheWind::KoalaRainbow::QueryParseNodes::Number;
use OnlyTheWind::KoalaRainbow::QueryParseNodes::QueryContext;
use OnlyTheWind::KoalaRainbow::QueryParseNodes::StringLiteral;
use OnlyTheWind::KoalaRainbow::QueryParseNodes::GlobalVariable;

=head1 Koala Rainbow Query Parser

This class implements the koala rainbow query parser.  In order to reduce
dependencies, and I suppose a bit of massochism and seeing if I can do it,
I am not using a helpful parser generator.  Actually, I suspect the whole
dependency concern might be moot, as it seems like the base Perl distribution
might include the recursive descent generator, but here we are.

I'm going to use regex's to perform the lexing, and just wail on the parsing
logic until it works.  Okay, so I went and looked at the dragon book, so now
I'm going to do things quasi-properly, match func and all.  It is appropriate
to note I have looked at the Perl XPath parser to see how it works.  It turns
out that combining standard OO principles with the dragon book and a language
to parse that's similar to XPath nets similar design (except theirs is better
:).

=over

=cut


sub new
{
    my $class = shift;
    my $self  =
    {
    };
    
    bless ($self, $class);
    return $self;
}

my $DEBUG_OFF = 0;
my $DEBUG_ERR = 1;
my $DEBUG_WARN = 2;
my $DEBUG_INFO = 3;
my $DEBUG_TRACE = 4;

my $debugLevel = $DEBUG_OFF;

sub _trace
{
    if($debugLevel >= $DEBUG_TRACE)
    {
        my $msg = shift;
        print $msg . "\n";
    }
}

sub _parseError
{
    my $self = shift;
    
    my $taggedTokens = $self->{TOKENS};
    my $idx          = $self->{IDX_CUR_TOKEN};

    my $subset = [];

    my $cur = 0;
    
    my $msg = '';
    
    for my $taggedToken (@$taggedTokens)
    {
        if($cur == $idx)
        {
            $msg .= '<|> ';
        }
        
        $msg .= $taggedToken->[1] . '(' . $taggedToken->[0]->[0] . ')';
    }
    continue
    {
        $cur++;
        
        $msg .= ' ';
    }
    
    $msg .= "\n";
        
    #print shift;
    #return undef;
    die $msg . (shift);
}

##### Lexing regex constant things #####
# Anything that induces recursive processing gets an OPEN and CLOSE constant.
# Everything else is just one-off.  Note that this is not the parse precedence
#  sequence, that happens in the lex func.
# It's not clear why I put those numbers there.  I guess I wanted a unique id,
#  which it certainly is, but that could be better accomplished through a helper
#  method or perhaps a variable that is evaluated and incremented each time.
#  Regardless, there is a number in the first field, that has no semantic value
#  other than uniqueness at this time.

# Arrr, slash be the hierarchy delimeter, ye salty dogs.
my $HIER_DELIM = [0, '/'];

# Constraints go between square brackets.
my $CONSTRAINT_OPEN  = [1, '\\['];
my $CONSTRAINT_CLOSE = [2, '\\]'];

# String constant.  Can be anything; doesn't stop till it comes out the other
#  side.
my $STRING_LITERAL = [3, "'([^']*)'"];

# Number, possibly stupid floating point.
my $NUMBER = [4, "(?:-)?\\d+(?:\\.\\d*)?"];

# Identifier.
my $IDENTIFIER = [5, '\w+'];

# Function-calls are indicated by parens, so is grouping.
my $PAREN_OPEN  = [6, '\\('];
my $PAREN_CLOSE = [7, '\\)'];

my $COMMA = [8, ','];

my $VAR_PREFIX = [9, '\\$'];

# Logical ops
my $LOG_AND = [10, 'and'];
my $LOG_OR  = [11, 'or'];

# Comparison ops
my $CMP_EQ  = [12, '='];
my $CMP_NE  = [13, '!='];
my $CMP_LT  = [14, '<'];
my $CMP_GT  = [15, '>'];
my $CMP_LTE = [16, '<='];
my $CMP_GTE = [17, '>='];

my $LEXER_PATTERNS = [$HIER_DELIM, $CONSTRAINT_OPEN, $CONSTRAINT_CLOSE,
                     $PAREN_OPEN, $PAREN_CLOSE, $COMMA, $VAR_PREFIX,
                     $LOG_AND, $LOG_OR,
                     $CMP_LTE, $CMP_GTE,, $CMP_EQ, $CMP_NE, $CMP_LT, $CMP_GT, 
                     $STRING_LITERAL, $NUMBER, $IDENTIFIER,];

my $CMP_OPS = [$CMP_LTE, $CMP_GTE, $CMP_EQ, $CMP_NE, $CMP_LT, $CMP_GT, ];

##### 

=item $parser->lex($text)

Lexes (tokenizes) the provided text.  There are basically three stages we go
through to do so (see below), with the end result providing us with a list of
lists, where each sub-list contains a reference to our lexing constants and
the string that is the token in question.

First, we create an aggregate regex out of all our patterns from above.  This is
sufficient to tokenize the input (#2), but we need to be able tell which regex
clause was fired.  It doesn't seem like there's actually a great way to do this.
Using $n sounds like a recipe for ugliness*, so my (according to me) less-ugly
solution is to then run all the individual regexes in order until we find who
the token matched against (#3).  Stupid?  Perhaps, but it saves me work and can
boost debugging by making it clear what we think each token mapped to, rather
than guessing.  *= I now recognize that a match performed in list context
without /g will return the captured values, but I need /g mode, and the list
context mode of /g doesn't seem to be what I want.  Actually, both might not
be right, because the intent would be to have many non-capturing values, whereas
the inclusion of only capturing values would foul things up.  This is definitely
an area for potential future optimizations.

=cut

sub lex
{
    my ($self, $text) = @_;
    
    # Transform the text for HTML reasons:
    $text =~ tr/\{/</;
    $text =~ tr/\}/>/;
    $text =~ s/&lt;/</g;
    $text =~ s/&gt;/>/g;
    
    # --- Build Custom Structures ---
    # We need to take the provided pattern list and use that to derive our own
    #  super-pattern, as well as per-pattern versions that also include the
    #  compiled version of the pattern.
    
    # The super pattern gobbles whitespace, with the nougat being a capture
    #  consisting of the or'ing together of all lexer patterns.
    my $superPattern = '\s*(';
    
    my $customPatterns = [];
    my $firstPass = 1;
    
    for my $lexPat (@$LEXER_PATTERNS)
    {
        my $pattern = $lexPat->[1];
    
        if(!$firstPass)
        {
            $superPattern .= '|';
        }
    
        # wrap the thing in a non-capturing grouping for complex cases.
        $superPattern .= '(?:' . $pattern . ')';
        
        my $customPat = [qr/$pattern/, $lexPat];
        
        push(@$customPatterns, $customPat);
        
        $firstPass = 0;
    }
       
    $superPattern .= ')\s*'; # gobble more whitespace.
    
    my $superMatcher = qr/$superPattern/;
    
    # --- Lex ---
    my $tokens = [];
    
    # Perform a match in global match mode in scalar context.  The code will
    #  loop until we stop matching.  Each match will start where the previous
    #  match ended thanks to the use of \G.  Because we use non-capturing
    #  groups, we need the 'x' option.  We use 's' because newlines should
    #  have no relevance to us.  We use 'c' because otherwise when our loop
    #  completes (by failing to match), the search position would be rest, and
    #  we would lose the ability to check it via the pos func.
    while($text =~ m/\G$superMatcher/gxsc)
    {
        # we will find out the type of the token in the next stage
        push(@$tokens, $1);
    }
    
    # --- Figure out what kind of token each token is ---
    my $taggedTokens = [];
    for my $token (@$tokens)
    {
        for my $customPat (@$customPatterns)
        {
            my $compiledPat = $customPat->[0];
            if($token =~ m/^$compiledPat$/)
            {
                #print 'Determined ' . $token . ' is ' .
                #      $customPat->[1]->[0] . "\n";
                #print '(matched ' . $compiledPat . ')' . "\n";
            
                my $lexPat = $customPat->[1];
                
                # if there was a grouping in the pattern, use it as the value
                #  (exclusively for string literals at this point, but this is
                #   a happy generalization.)
                if($1)
                {
                    $token = $1;
                }
                
                push(@$taggedTokens, [$lexPat, $token]);
                last;
            }
        }
    }
    
    return $taggedTokens;
}

sub _stringify_tagged_tokens
{
   my $taggedTokens = shift;
   
   my $str = '';
   for my $taggedToken (@$taggedTokens)
   {
       $str .= $taggedToken->[1] . "   ";
   }
   
   return $str;
}

sub _remainingTokens
{
    my $self = shift;
    
    my $taggedTokens = $self->{TOKENS};
    my $idx          = $self->{IDX_CUR_TOKEN};

    my $subset = [];

    my $cur = 0;
    for my $taggedToken (@$taggedTokens)
    {
        if($cur >= $idx)
        {
            push(@$subset, $taggedToken);
        }
    }
    continue
    {
        $cur++;
    }    

    return $subset;

    #print 'my slice range is: ' . $idx . ' ' . @$tokens;
    #print 'my slice is ' . join(' ',  @{$tokens->[$idx..@$tokens]}) . "\n";
    #return $tokens->[$idx..@$tokens];
}

=item $parser->parseSetQuery()

Initiates a parse of a set query.  A set query's top-level is only capable of
1) an optional global set reference followed by edge-traversals, where each
edge traversal may contain constraints; 2) a function call; 3) a variable
reference.  The constraints can do prettymuch anything as long as each results
in a boolean value; they are equivalent to calls to parseBooleanQuery.

=cut

sub parseSetQuery
{
    my ($self, $text) = @_;
    
    my $tokens = $self->lex($text);
    
    $self->{TOKENS} = $tokens;
    $self->{IDX_CUR_TOKEN} = 0;
    
    my $tree = $self->_parseSetBottomExpr(0, 1);
    
    return $tree;
}

=item $parser->parseValueQuery()

Initiates a parse of a value query, whose evaluated result will be a non-boolean
value.  (Use parseBooleanQuery if you want a boolean output.)  A value query
is basically the same as a set query, but with the ability to directly return
a numeric constant or string literal.

=cut

sub parseValueQuery
{
    my ($self, $text) = @_;
    
    my $tokens = $self->lex($text);
    
    $self->{TOKENS} = $tokens;
    $self->{IDX_CUR_TOKEN} = 0;
    
    my $tree = $self->_parseValueBottomExpr(1);
    
    return $tree;
}

=item $parser->parseBooleanQuery()

Initiates a parse of a boolean query, whose evaluated result will be a boolean
value.  This call is equivalent to what happens in a set parse constraint block.
Comparisons and logical operations are valid at a top-level here.  Of course,
any nested constraint blocks obey their standard rules.

=cut

sub parseBooleanQuery
{
    my ($self, $text) = @_;
    
    my $tokens = $self->lex($text);
    
    $self->{TOKENS} = $tokens;
    $self->{IDX_CUR_TOKEN} = 0;
    
    my $tree = $self->_parseLogicalOrExpr(1);
    
    return $tree;
}

sub _peek
{
    my $self = shift;
    my $tag  = shift;
    my $ahead = shift;
    if(!$ahead)
    {
        $ahead = 0;
    }
    
    my $tokens = $self->{TOKENS};
    my $idx    = $self->{IDX_CUR_TOKEN};

    if($idx + $ahead >= @$tokens)
    {
        return 0;
    }    
    
    if($tokens->[$idx + $ahead]->[0] == $tag)
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

sub _peekAll
{
    my $self = shift;
    my $tags  = shift;
    my $ahead = shift or 0;
    
    my $tokens = $self->{TOKENS};
    my $idx    = $self->{IDX_CUR_TOKEN};

    if($idx + $ahead >= @$tokens)
    {
        return 0;
    }    
        
    for my $tag (@$tags)
    {
        if($tokens->[$idx + $ahead]->[0] == $tag)
        {
            return 1;
        }
    }
    
    return 0;
}

sub _match
{
    my ($self, $tag) = @_;
    my $tokens = $self->{TOKENS};
    my $idx    = $self->{IDX_CUR_TOKEN};
    
    if($idx >= @$tokens)
    {
        return 0;
    }    
    
    #print 'Comparing' . $tokens->[$idx]->[0] . ' with ' . $tag . "\n";
    if($tokens->[$idx]->[0] == $tag)
    {
        #print "match!\n";
        # matched, gobble the token by incrementing our counter.
        $self->{IDX_CUR_TOKEN} = $idx + 1;
        
        return 1;
    }
    else
    {
        return 0;
    }
}

sub _matchAll
{
    my ($self, $tags) = @_;
    my $tokens = $self->{TOKENS};
    my $idx    = $self->{IDX_CUR_TOKEN};

    if($idx >= @$tokens)
    {
        return 0;
    }
    
    for my $tag (@$tags)
    {
        if($tokens->[$idx]->[0] == $tag)
        {
            # matched, gobble the token by incrementing our counter.
            $self->{IDX_CUR_TOKEN} = $idx + 1;
            return 1;
        }
    }
    
    return 0;
}

sub _getLastMatch
{
    my $self = shift;
    my $tokens = $self->{TOKENS};
    my $idx    = $self->{IDX_CUR_TOKEN};
    
    return $tokens->[$idx - 1]->[1];
}

sub _noMoreTokens
{
    my $self = shift;
    
    my $tokens = $self->{TOKENS};
    my $idx    = $self->{IDX_CUR_TOKEN};
    
    return ($idx >= @$tokens);
}

sub _parseGlobalContext
{
    my ($self, $constraintMode) = @_;

    _trace('GlobalContext');

    # If we see a hierarchy delimiter here, that means the input is from a
    #  global-space.  This is only appropriate for the base-case, which means
    #  the very beginning of a query parse, or inside a constraint.
    # In this case, the only legal scenario is then another hier delim, or
    #  a constraint.
    if($self->_match($HIER_DELIM))
    {
        my $context =
            OnlyTheWind::KoalaRainbow::QueryParseNodes::QueryContext->new();

        # also, the next token must be an identifier
        if($self->_match($IDENTIFIER))
        {
            $context->globalSet($self->_getLastMatch());
        }
        else
        {
            return $self->_parseError('Expected identifier following "/" indicating' .
                               ' global scope!');
        }
        
        # is there a constraint?
        if($self->_match($CONSTRAINT_OPEN))
        {
            my $constraint = $self->_parseLogicalOrExpr(1);
            
            $context->constraint($constraint);
            
            if(!$self->_match($CONSTRAINT_CLOSE))
            {
                return $self->_parseError('Expected bracket closing constraint.');
            }
        }
         
        return $context;
    }

    # there was no global context def.
    return undef;
}

# Handle:
# [blah] (/ ...)
# func(blah) ([blah]) (/ ...)
# edge ([blah]) (/ ...)
# 
sub _parseContext
{
    my ($self, $constraintMode, $top) = @_;
    
    _trace('Context');
    
    my $context =
        OnlyTheWind::KoalaRainbow::QueryParseNodes::QueryContext->new();

    my $evalNode;
    
    if($constraintMode)
    {
        # logical and comparison ops are valid
        $evalNode = $self->_parseLogicalOrExpr($constraintMode);
    }
    else
    {
        # only value returning people are valid; use parseBottomExpr
        $evalNode = $self->_parseBottomExpr($constraintMode);
    }
    
    $context->evalNode($evalNode);

    # there may or may not be a constraint
    if($self->_match($CONSTRAINT_OPEN))
    {
        my $constraint = $self->_parseNewContext(1, 0);
        
        if(!$self->_match($CONSTRAINT_CLOSE))
        {
            return $self->_parseError('Expected constraint closing brace.');
        }
        
        $context->constraint($constraint);
    }
    
    # there may or may not be a /
    if($self->_match($HIER_DELIM))
    {
        my $nextContext = $self->_parseContext(0, 0);
        
        $context->nextContext($nextContext);
        
        return $context;
    }
    # if we are a top-level parse, there should be no more tokens
    elsif($top and !$self->_noMoreTokens())
    {
        
        return $self->_parseError('Expected to have no more tokens! But some ' .
                           "remain: \n" .
                           _stringify_tagged_tokens($self->_remainingTokens()));
    }
    
   return $context;
}

# Okay, so this is our top-level expression parsing jig.  Our precedence
#  is basically (logical, comparison, func/constant/subexpr/edge), where logical
#  is in turn broken down into (or, and).  Comparisons don't need to be further
#  subdivided because we they take in scalars and output boolean, so their
#  outputs can't be chained.  Subexpr just means some parens that aren't a
#  function call and thus start again at parseExpr!
#sub parseExpr
#{
#    my $self = shift;
#}

# Handle:
#  blah
#  blah or blah
#  blah or blah or blah ...
sub _parseLogicalOrExpr
{
    my ($self, $constraintMode) = @_;
 
    _trace("LogicalOrExpr");
    
    my $expr = $self->_parseLogicalAndExpr($constraintMode);
    while($self->_match($LOG_OR))
    {
        my $left = $expr;
        $expr =
            OnlyTheWind::KoalaRainbow::QueryParseNodes::Expression->new();
        $expr->left($left);
        
        $expr->op('or');
        
        $expr->right($self->_parseLogicalAndExpr($constraintMode));
    }
    
    return $expr;
}

# same deal
sub _parseLogicalAndExpr
{
    my ($self, $constraintMode) = @_;

    _trace("LogicalAndExpr");
    
    my $expr = $self->_parseComparisonExpr($constraintMode);
    #######
    _trace(_stringify_tagged_tokens($self->_remainingTokens()));
    while($self->_match($LOG_AND))
    {
        _trace('  and');
    
        my $left = $expr;
        $expr =
            OnlyTheWind::KoalaRainbow::QueryParseNodes::Expression->new();
        $expr->left($left);
        
        $expr->op('and');
        
        $expr->right($self->_parseComparisonExpr($constraintMode));
    }
    
    return $expr;    
}

# Boolean returning!
# Handle
# blah
# blah OP blah
sub _parseComparisonExpr
{
    my ($self, $constraintMode) = @_;

    _trace("ComparisonExpr");
    
    my $expr = $self->_parseConstraintBottomExpr($constraintMode);
    if($self->_matchAll($CMP_OPS))
    {
        my $left = $expr;
        $expr =
            OnlyTheWind::KoalaRainbow::QueryParseNodes::Expression->new();
        $expr->left($left);
        
        $expr->op($self->_getLastMatch());
        
        $expr->right($self->_parseConstraintBottomExpr($constraintMode));
    }
    
    return $expr;    
}

# Handle:
# (expr)
# -- chain to value bottom --
sub _parseConstraintBottomExpr
{
    my ($self, $constraintMode) = @_;
    
    _trace('ConstraintBottom');
    
    if($self->_match($PAREN_OPEN))
    {
        my $expr = $self->_parseLogicalOrExpr($constraintMode);
        
        if(!$self->_match($PAREN_CLOSE))
        {
            return $self->_parseError('Expected closing paren for grouping!');
        }
        
        return $expr;
    }
    
    return $self->_parseValueBottomExpr($constraintMode);
}

# Handle:
# 4
# 'Hello!'
# -- chain to set bottom --
sub _parseValueBottomExpr
{
    my ($self, $constraintMode) = @_;
    
    _trace('ValueBottom');
    
    if($self->_match($NUMBER))
    {
        my $val = $self->_getLastMatch();
        return OnlyTheWind::KoalaRainbow::QueryParseNodes::Number->new($val);
    }
    elsif($self->_match($STRING_LITERAL))
    {
        my $val = $self->_getLastMatch();
        return OnlyTheWind::KoalaRainbow::QueryParseNodes::StringLiteral->new(
                   $val);
    }
    
    return $self->_parseSetBottomExpr($constraintMode);
}

# Handle:
# $var
# -- chain to context set bottom --
sub _parseSetBottomExpr
{
    my ($self, $constraintMode) = @_;
    
    _trace('SetBottom');

    return $self->_parseContextSetBottomExpr($constraintMode);
}

# This is broken into its own set because an edge-traversal chain could
#  conceivably, at any point, be made up of an edge traversal or a function
#  call that relies on the context at that stage of the edge traversal.
# Handle:
# foo(newcontext, newcontext, newcontext, ...)
# edge
#   prefix: /globalset/
# suffixes for all:
#  [constraint]
#  /(edgetraversal)
sub _parseContextSetBottomExpr
{
    my ($self, $constraintMode) = @_;
    
    _trace('ContextSetBottom');

    my $context =
        OnlyTheWind::KoalaRainbow::QueryParseNodes::QueryContext->new();
    my $retcontext = $context;

    do
    {
        # variable reference
        if($self->_match($VAR_PREFIX))
        {
            if(!$self->_match($IDENTIFIER))
            {
                return $self->_parseError('Expected identifier after $!');
            }
            
            my $globalVar =
                OnlyTheWind::KoalaRainbow::QueryParseNodes::GlobalVariable->new();
            
            $globalVar->name($self->_getLastMatch());
            
            $context->evalNode($globalVar);
        }
        # catch the function call case (identifier followed by paren)
        elsif($self->_peek($PAREN_OPEN, 1) and
              $self->_match($IDENTIFIER))
        {
            my $funcCall =
                OnlyTheWind::KoalaRainbow::QueryParseNodes::FunctionCall->new();
            
            $funcCall->name($self->_getLastMatch());
            
            # eat open paren, know it matches
            $self->_match($PAREN_OPEN);
        
            my $args = [];
        
            if(!$self->_peek($PAREN_CLOSE))
            {
                do
                {
                    # arguments can be set, value, or constraint.
                    my $arg = $self->_parseLogicalOrExpr(1);
                    push(@$args, $arg);
                }
                while($self->_match($COMMA));
            }
            
            $funcCall->args($args);
            
            # ran out of commas, there better be a right paren here...
            if(!$self->_match($PAREN_CLOSE))
            {
                return $self->_parseError('Expected closing paren for function!');
            }
            
            $context->evalNode($funcCall);
        }
        # handle the /globalset case
        elsif($self->_match($HIER_DELIM))
        {
            # also, the next token must be an identifier
            if($self->_match($IDENTIFIER))
            {
                $context->globalSet($self->_getLastMatch());
            }
            else
            {
                return $self->_parseError('Expected identifier following "/" '.
                                          ' indicating global scope!');
            }
        }        
        elsif($self->_match($IDENTIFIER))
        {
            my $contextVar =
             OnlyTheWind::KoalaRainbow::QueryParseNodes::ContextVariable->new();
            
            $contextVar->name($self->_getLastMatch());
            
            $context->evalNode($contextVar);
        }

        # is there a constraint?
        if($self->_match($CONSTRAINT_OPEN))
        {
            my $constraint = $self->_parseLogicalOrExpr(1);
            
            $context->constraint($constraint);
            
            if(!$self->_match($CONSTRAINT_CLOSE))
            {
                return $self->_parseError('Expected bracket closing constraint.');
            }
        }

        # this method isn't exactly my crowning achievement, so this horrible
        #  hack won't cost me any sleep.
        if($self->_peek($HIER_DELIM))
        {
            my $nextContext =
                OnlyTheWind::KoalaRainbow::QueryParseNodes::QueryContext->new();
    
            $context->nextContext($nextContext);
            $context = $nextContext;
        }
    }
    while($self->_match($HIER_DELIM));
    
    return $retcontext;
}

=pod

=back

=cut

1;