#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::Util;
use strict;

use MT::Util qw (days_in start_end_day start_end_week start_end_month 
                 offset_time_list offset_time );

use Time::Local qw( timegm );

use Exporter;
@OnlyTheWind::KoalaRainbow::Util::ISA = qw( Exporter );

use vars qw( @EXPORT_OK );
@EXPORT_OK = qw( applyTimeStampDelta parseTimeInterval thresholdTimeStamp
                 nowTimeStamp secsToTimeStamp timeStampToSecs
                 reverseEngineerFileInfo
               );

# This method takes a Template::Context and attempts to determine the type of
#  file it is: index, or archive (individual, daily, weekly, monthly, category).
#  Because I'm hoping that we can use archive_file_for's logic to tell us
#  filenames, our returned type values match those used by archive_file_for,
#  and presumably the rest of the MT system.
sub reverseEngineerFileInfo
{
    my ($context) = @_;
    
    my $type = '';
    
    my $start = $context->{current_timestamp};
    my $end = $context->{current_timestamp_end};
    
    if(!defined($end))
    {
        # if it's an archive, it's an individual
        $type = 'Individual';
    }
    else
    {
        # Same day?
        if(substr($start, 0, 8) eq substr($end, 0, 8))
        {
            $type = 'Daily';
        }
        # Same week? (it would be 7 days less 1 second later.
        if(applyTimeStampDelta($start, '7d-1s') eq $end)
        {
            $type = 'Weekly';        
        }
        # Same month?
        elsif(substr($start, 0, 6) eq substr($end, 0, 6))
        {
        }
    }
}

sub timeStampToSecs
{
    my ($blog, $timestamp) = @_;

    my($year, $month, $day, $hour, $min, $sec) = 
        unpack 'A4A2A2A2A2A2', $timestamp;    
        
    # although we fix the month offset, the Time::Local documentation states for
    #  the year that values >=1000 are used verbatim, whereas it has some
    #  tricky logic in the other cases.  Strictly speaking, given the year 2038
    #  problem, they're all equally good given the current date, but this one
    #  will perhaps work for hundreds of years longer assuming the seconds
    #  since epoch thing is fixed via 64-bit int or some such.
    # NOTE: timegm presumably thinks we are giving it a GM-centric date/time.
    #  Since we use the blog-offset time as our wall-clock, which may not be
    #  the same as the local time zone, we can't use timelocal.  Although we
    #  will pass the number of seconds through the blog offset reversal, this
    #  presumably will not be perfect as this will happen after Time::Local does
    #  its DST and leap magic.  Honestly, I don't care to think about it at
    #  all right now, so we're basically just doing the inverse of what we do
    #  in nowTimeStamp, which is sufficient for my purposes.
    my $secs = timegm($sec, $min, $hour, $day, $month-1, $year);
    
    return offset_time($secs, $blog, '-');
}

sub secsToTimeStamp
{
    my $blog = shift;
    my $secs = shift;
    
    # this invokes gmtime on the transformed epoch-secound-count...
    # note that the year returned is 1900-relative, the month is 0-based, but
    #  mday is oddly 1-based.  (perlfun only implies it, but time.h comes out
    #  and says it.)
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) =
        offset_time_list($secs, $blog);
    
    my $timestamp = sprintf("%04d%02d%02d%02d%02d%02d",
                            $year+1900, $mon+1, $mday, $hour, $min, $sec);

    return $timestamp;    
}

sub nowTimeStamp
{
    my $blog = shift;
    my $now = time;

    return secsToTimeStamp($blog, $now);
}

# w: week
# d: day
# h: hour
# m: minute
# s: sec
sub parseTimeInterval
{
    my $text = shift;
    my @r = $text =~
        m/^(?:(-?\d+)y)?(?:(-?\d+)mo)?(?:(-?\d+)w)?(?:(-?\d+)d)?(?:(-?\d+)h)?(?:(-?\d+)m)?(?:(-?\d+)s)?$/;
    return @r;        
}

# deltas are applied from second up through year (this is relevant when the
#  delta tweaks more than just one 'field'.
# month just goes to the same day in the next month.  if that month lacks that
#  day (29 of feb, 31 of jan.)
sub applyTimeStampDelta
{
    my ($timestamp, $delta) = @_;
    
#    use Data::Dump qw(dump);
#    print "parsed as: " . dump(parseTimeInterval($delta));
    
    my ($dyear, $dmonth, $dweek, $dday, $dhour, $dmin, $dsec) =
        parseTimeInterval($delta);
    
    my($year, $month, $day, $hour, $min, $sec) = 
        unpack 'A4A2A2A2A2A2', $timestamp;

    if($dweek)
    {
        $dday += $dweek * 7;
    }

    if($dsec)
    {
        $sec += $dsec;
        if($sec >= 60)
        {
            $min += int($sec / 60);
            $sec %= 60;
        }
        elsif($sec < 0)
        {
            # the rhs is negative, so this is truly subtraction
            $min += int(($sec - 59) / 60);
            # this does the right thing (goes positive)
            $sec %= 60;
        }
    }
    
    if($dmin)
    {
        $min += $dmin;
    }
    if($min >= 60)
    {
        $hour += int($min / 60);
        $min %= 60;
    }
    elsif($min < 0)
    {
        $hour += int(($min - 59) / 60);
        $min %= 60;
    }
    
    if($dhour)
    {
        $hour += $dhour;
    }
    if($hour >= 24)
    {
        $day += int($hour / 24);
        $hour %= 24;
    }
    elsif($hour < 0)
    {
        $day += int(($hour - 23) / 24);
        $hour %= 24;
    }

    fixday:
    if($dday)
    {
        $day += $dday;
    }
    while($day > days_in($month, $year))
    {
        my $dim = days_in($month, $year);
        $day -= $dim;
        $month += 1;
        
        if($month > 12)
        {
            $year += 1;
            $month = 1;
        }
    }
    while($day < 1)
    {
        if($month == 1)
        {
            $month = 12;
            $year -= 1;
        }
        else
        {
            $month -= 1;
        }
        my $dipm = days_in($month, $year);

        $day += $dipm;
    }
    
    if($dmonth)
    {
        $month += $dmonth;
    }
    if($month > 12)
    {
        $year += int(($month - 1)/ 12);
        $month = (($month - 1) % 12) + 1;
    }
    # okay, this one would be too confusing without a loop. stupid 0/1 issue.
    # just spin, adding 12 until the months are in the right range again.
    while($month < 1)
    {
        $year -= 1;
        $month += 12;
    }
    # so, uh, the month change might have changed the number of days in the
    #  month.  I've never been so happy to see a goto command!
    goto fixday if($day > days_in($month, $year));    
    
    if($dyear)
    {
        $year += $dyear;
    }
    # so, uh, the year change might have changed the number of days in the
    #  month...
    goto fixday if($day > days_in($month, $year));    

    return sprintf("%04d%02d%02d%02d%02d%02d",
                   $year, $month, $day,
                   $hour, $min, $sec);
}

# basically, selective floor() for dates.  The assumption is that people are
#  not going to ever want to round up (for the purposes) we have conceived.
#  What is more likely and more intuitive is that they would want to threshold,
#  then _ADD_ a day/week/month, whatever, to the thresholded date.  This is
#  more intuitive because our behavior is equivalent to always telling you
#  the today/this week/this month/this year.  Tomorrow is then relative to it.
#  Trying to let the thresholding do ceil() as well would be complex and
#  confusing.
sub thresholdTimeStamp
{
    my ($timestamp, $threshold) = @_;
    
    if($threshold eq '')
    {
        return $timestamp;
    }
    elsif($threshold eq 'y')
    {
        # quote the year, the first day is jan 1st, at 00:00:00.
        return substr($timestamp, 0, 4) . '0101' . '000000';
    }
    elsif($threshold eq 'mo')
    {
        return scalar(start_end_month($timestamp));
    }
    elsif($threshold eq 'w')
    {
        return scalar(start_end_week($timestamp));
    }
    elsif($threshold eq 'd')
    {
        return scalar(start_end_day($timestamp));
    }
    elsif($threshold eq 'h')
    {
        # drop the min:sec, replace with 00:00
        return substr($timestamp, 0, 10) . '0000';
    }
    elsif($threshold eq 'm')
    {
        # drop the sec
        return substr($timestamp, 0, 12) . '00';
    }
    else
    {
        return $timestamp;
    }
}

1;