#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::Visualization;
use strict;

use OnlyTheWind::KoalaRainbow::ContextManager;
use OnlyTheWind::KoalaRainbow::Drawing;
use OnlyTheWind::KoalaRainbow::Environment;
use OnlyTheWind::KoalaRainbow::QueryParser;

use OnlyTheWind::KoalaRainbow::DrawNodes::Rectangle;
use OnlyTheWind::KoalaRainbow::DrawNodes::Ellipse;
use OnlyTheWind::KoalaRainbow::DrawNodes::Line;

use MT::Util qw(format_ts);

# This class encapsulates most/all of the actual plugin tag processing.
# We're talking context stack maintenance, parsing of arguments, etc.
# The goal at this point is to have a working implementation, refactoring
#  will come later (or not at all ;).

sub new
{
    my $topActionList = [];

    my $class = shift;
    my $self  =
    {
        DRAWING  => OnlyTheWind::KoalaRainbow::Drawing->new(),
        PARSER => OnlyTheWind::KoalaRainbow::QueryParser->new(),
        NULL_CONTEXT => OnlyTheWind::KoalaRainbow::ContextManager->new(),
        ERR_COUNT => 0,
        ERR_STR => '',
        ROOT_PATH => undef,
        BASE_DIR => undef,
        FILENAME => undef,
        CONTEXT_STACK => [],
        ACTION_LIST => $topActionList,
        ACTION_LIST_STACK => [$topActionList],
    };
    
    bless ($self, $class);
    
    $self->_pushContext($self->{NULL_CONTEXT});
    
    return $self;
}

sub _logErr
{
    my $self = shift;
    my $msg = shift;

    $self->{ERR_COUNT} = $self->{ERR_COUNT} + 1;
    $self->{ERR_STR} = $self->{ERR_STR} . $msg . "\n";

    return undef;  
}

sub errorCount
{
    my $self = shift;
    return $self->{ERR_COUNT};
}

sub errorMessages
{
    my $self = shift;
    return $self->{ERR_STR};
}

sub _parseSet
{
    my $self = shift;
    my ($queryText, $ctxMgr) = @_;
    my $parser = $self->{PARSER};

    my $tree = undef;

    eval
    {
        local $SIG{'__DIE__'};
        $tree = $parser->parseSetQuery($queryText);
    };
    # problem?
    if($@)
    {
        $self->_logErr('Parse Error: ' . $@);
        return undef;
    }
    
    return $tree;
}

sub _evalSet
{
    my $self = shift;
    my ($tree, $ctxMgr) = @_;
    
    my $resCtxMgr;
    
    eval
    {
        local $SIG{'__DIE__'};
        $tree = $tree->clone();
        $tree = $tree->partialEvaluate($ctxMgr);
        $resCtxMgr = $tree->evaluateSet($ctxMgr);
    };
    
    if($@)
    {
        $self->_logErr('Eval Error: ' . $@);
        return undef;
    }
    else
    {
        return $resCtxMgr;
    }
}

sub _envSetSet
{
    my $self = shift;
    my ($name, $val) = @_;
    
    return OnlyTheWind::KoalaRainbow::Environment->setset($name, $val);
}

sub _parseAndEvalValue
{
    my $self = shift;
    my ($queryText, $ctxMgr) = @_;

    my $tree = $self->_parseValue($queryText, $ctxMgr);
    return $self->_evalValue($tree, $ctxMgr);
}

sub _parseValue
{
    my $self = shift;
    my ($queryText, $ctxMgr) = @_;
    my $parser = $self->{PARSER};

    my $tree = undef;

    eval
    {
        local $SIG{'__DIE__'};
        $tree = $parser->parseValueQuery($queryText);
    };
    # problem?
    if($@)
    {
        $self->_logErr('Parse Error: ' . $@);
        return undef;
    }
    
    return $tree;
}

sub _evalValue
{
    my $self = shift;
    my ($tree, $ctxMgr) = @_;

    my $val;
    
    eval
    {
        local $SIG{'__DIE__'};
        $tree = $tree->clone();
        $tree = $tree->partialEvaluate($ctxMgr);
        $val = $tree->evaluateValue($ctxMgr);
    };
    
    if($@)
    {
        #die $@;
        $self->_logErr('Eval Error: ' . $@);
        return 0;
    }
    elsif(!defined($val))
    {
        $self->_logErr('Evaluation of ' . $tree->toString() . ' returned an ' .
                       'undefined value.');
        return 0;
    }
    else        
    {
        return $val;
    }
}

sub _envSetVal
{
    my $self = shift;
    my ($name, $val) = @_;
    
    return OnlyTheWind::KoalaRainbow::Environment->setval($name, $val);
}

sub _parseBoolean
{
    my $self = shift;
    my ($queryText, $ctxMgr) = @_;
    my $parser = $self->{PARSER};

    my $tree = undef;

    eval
    {
        local $SIG{'__DIE__'};
        $tree = $parser->parseBooleanQuery($queryText);
    };
    # problem?
    if($@)
    {
        $self->_logErr('Parse Error: ' . $@);
        return undef;
    }
    
    return $tree;
}

sub _evalBoolean
{
    my $self = shift;
    my ($tree, $ctxMgr) = @_;

    my $val;
    
    eval
    {
        local $SIG{'__DIE__'};
        $tree = $tree->clone();
        $tree = $tree->partialEvaluate($ctxMgr);
        $val = $tree->evaluateValue($ctxMgr);
    };
    
    if($@)
    {
        #die $@;
        $self->_logErr('Eval Error: ' . $@);
        return 0;
    }
    elsif(!defined($val))
    {
        $self->_logErr('Evaluation of ' . $tree->toString() . ' returned an ' .
                       'undefined value.');
        return 0;
    }
    else        
    {
        return $val;
    }
}

sub _pushContext
{
    my $self = shift;
    my ($context) = @_;
    
    push(@{$self->{CONTEXT_STACK}}, $context);
}

sub _popContext
{
    my $self = shift;
    
    return pop(@{$self->{CONTEXT_STACK}});
}

sub _peekContext
{
    my $self = shift;
    
    return $self->{CONTEXT_STACK}->[-1];
}

sub _procFilename
{
    my ($self, $context, $args) = @_;

    if($context->stash('KRoutputfilepathoverride'))
    {
        return;
    }

    my $filename_to_date_proc;

    if(defined($args->{filename_dyn}))
    {
        $filename_to_date_proc = $self->_parseAndEvalValue($args->{filename_dyn},
                                                   $self->_peekContext());
    }
    elsif(defined($args->{filename}))
    {
        $filename_to_date_proc = $args->{filename};
    }
    else
    {
        $self->_logErr('No filename or filename_dyn attribute defined.');
        return;
    }

    my $filename;

    if(defined $context->{current_timestamp})
    {
        $filename = format_ts($filename_to_date_proc,
                              $context->{current_timestamp},
                              $context->stash('blog'));
    }
    else
    {
        $filename = $filename_to_date_proc;
    }

    $self->{FILENAME} = $filename;
}

sub procVisualization
{
    my $self = shift;
    my ($context, $args) = @_;
    
    my $nullCtxMgr = $self->{NULL_CONTEXT};
    $nullCtxMgr->hint('mtcontext', $context);
    
    my $drawing = $self->{DRAWING};

    if(exists $args->{width})
    {
        my $width;
        if($args->{width} eq 'auto')
        {
            $width = 'auto';
        }
        else
        {
            $width = $self->_parseAndEvalValue($args->{width}, $nullCtxMgr);
        }
        $drawing->width($width);
    }
    
    if(exists $args->{height})
    {
        my $height;
        if($args->{height} eq 'auto')
        {
            $height = 'auto';
        }
        else
        {
            $height = $self->_parseAndEvalValue($args->{height}, $nullCtxMgr);
        }
        $drawing->height($height);
    }
    
    if(exists $args->{imagemap})
    {
        $drawing->imageMapName($args->{imagemap});
    }
    
    if(exists $args->{bg})
    {
        my $background = $self->_parseAndEvalValue($args->{bg}, $nullCtxMgr);
        $drawing->background($background);
    }
    
    if(exists $args->{scale})
    {
        if($args->{scale} eq 'auto')
        {
            $drawing->autoScale(1);
        }
        elsif($args->{scale} eq 'auto-aspect')
        {
            $drawing->autoScale(2);
        }
        if(defined($args->{scalelayer}))
        {
            $drawing->scaleLayer($args->{scalelayer});
        }
        
        my ($padLeft, $padTop, $padRight, $padBottom) = (0, 0, 0, 0);
        if($args->{padall})
        {
            $padLeft = $padTop = $padRight = $padBottom = $args->{padall}; 
        }
        if($args->{padleft})
        {
            $padLeft = $args->{padleft};
        }
        if($args->{padtop})
        {
            $padTop = $args->{padtop};
        }
        if($args->{padright})
        {
            $padRight = $args->{padright};
        }
        if($args->{padbottom})
        {
            $padBottom = $args->{padbottom};
        }
        $drawing->pad($padLeft, $padTop, $padRight, $padBottom);
    }
    
    # filename!
    $self->_procFilename($context, $args);
    
    1;
}

sub filename
{
    my $self = shift;
    
    if (@_)
    {
        $self->{FILENAME} = shift;
    }
    
    return $self->{FILENAME};
}

sub _determineSavePath
{
    my $self = shift;
    my ($context) = @_;
    
    if($context->stash('KRoutputfilepathoverride'))
    {
        return $context->stash('KRoutputfilepathoverride');
    }
    else
    {
        my $blog = $context->stash('blog');
    
        my $outputFilePath = File::Spec->catfile($blog->site_path,
                                                 $self->{FILENAME});
        return $outputFilePath;
    }
}

sub _determineURLPath
{
    my $self = shift;
    my ($context) = @_;
    
    if($context->stash('KRoutputurlpathoverride'))
    {
        return $context->stash('KRoutputurlpathoverride');
    }
    else
    {
        my $blog = $context->stash('blog');
    
        my $outputURLPath = $blog->site_url;
        if(!$outputURLPath =~ m!/$!)
        {
            $outputURLPath .= '/';
        }
        $outputURLPath .= $self->{FILENAME};   
        
        return $outputURLPath;
    }
}

sub render
{
    my $self = shift;
    my ($context) = @_;
    my $drawing = $self->{DRAWING};

    my $imageData;

    eval
    {
        local $SIG{'__DIE__'};
        $imageData = $drawing->draw();
    };

    if($@)
    {
        $self->_logErr('Exception when drawing: ' . $@);
        return undef;
    }

    my $savepath = $self->_determineSavePath($context);
    
    my $blog = $context->stash('blog');
    my $fileMgr = $blog->file_mgr;
    
#die 'want to write to ' . $savepath;

#use Data::Dump qw(dump);
#die 'data: ' . dump($imageData);
    
    $fileMgr->put_data($imageData, $savepath, 'upload');

    return 1;
}

sub getHTMLSnippet
{
    my $self = shift;
    my ($context) = @_;
    
    my $drawing = $self->{DRAWING};
    
    my $imageurl = $self->_determineURLPath($context);
    
    my $str = '';
    
    my $imageMapRef = '';
    if(defined($drawing->imageMapName))
    {
        $str .= $drawing->imageMapContents . "\n";
        $imageMapRef = ' usemap="#' . $drawing->imageMapName . '" border="0"';
    }
    
    $str .= '<img src="' . $imageurl . '" ' .
             'width="' . $drawing->width() . '" ' .
             'height="' . $drawing->height() . '"' .
             $imageMapRef . '>';
}

sub _addAction
{
    my $self = shift;
    my ($nodeName, $nodeVal) = @_;
    my $curActionList = $self->{ACTION_LIST_STACK}->[-1];
    push(@$curActionList, [$nodeName, $nodeVal]);
}

sub _parseHashAsValues
{
    my $self = shift;
    my $inHash = shift;

    my $resSet = {};

    while (my ($key, $value) = each %$inHash)
    {
        $resSet->{$key} = $self->_parseValue($value);
    }
    
    return $resSet;
}

sub _parseHashAsBooleans
{
    my $self = shift;
    my $inHash = shift;

    my $resSet = {};

    while (my ($key, $value) = each %$inHash)
    {
        $resSet->{$key} = $self->_parseBoolean($value);
    }
    
    return $resSet;
}

sub _parseHashAsSets
{
    my $self = shift;
    my $inHash = shift;

    my $resSet = {};

#    my ($key, $value);

    while (my ($key, $value) = each %$inHash)
    {
        $resSet->{$key} = $self->_parseSet($value);
    }
        
    return $resSet;
}

sub _pushActionList
{
    my $self = shift;
    my ($actionList) = @_;
    
    push(@{$self->{ACTION_LIST_STACK}}, $actionList);
}

sub _popActionList
{
    my $self = shift;
    
    return pop(@{$self->{ACTION_LIST_STACK}});
}

sub _runList
{
    my $self = shift;
    my ($runList) = @_;
    
    for my $runItem (@$runList)
    {
        my ($nodeName, $nodeVal) = @$runItem;
    
        if ($nodeName eq 'if')
        {
            $self->procIf($nodeVal);
        }
        elsif($nodeName eq 'loop')
        {
            $self->procLoop($nodeVal);
        }
        elsif($nodeName eq 'setset')
        {
            $self->procSetSets($nodeVal);
        }
        elsif($nodeName eq 'setval')
        {
            $self->procSetVals($nodeVal);
        }
        elsif($nodeName eq 'line')
        {
            $self->procLine($nodeVal);
        }
        elsif($nodeName eq 'rect')
        {
            $self->procRect($nodeVal);
        }
        elsif($nodeName eq 'ellipse')
        {
            $self->procEllipse($nodeVal);
        }
        elsif($nodeName eq 'filter')
        {
            $self->procFilter($nodeVal);
        }
    }
}

sub run
{
    my $self = shift;
    
    $self->_runList($self->{ACTION_LIST});
}

sub procPrint
{
    my $self = shift;
    my ($context, $args) = @_;

    for my $key (keys %$args)
    {
        my $query = $args->{$key};
        my $val = $self->_evalValue($query, $self->_peekContext());
        warn $key . ': ' . $val;
    }
    
    return 1;
}

sub cmdSetSets
{
    my $self = shift;
    $self->_addAction('setset', $self->_parseHashAsSets(shift));
}

sub procSetSets
{
    my $self = shift;
    my ($args) = @_;

    # in order to avoid ugly behavior, we calculate all the answers, THEN
    #  we apply them to the env.
    my $answers = {};

    for my $key (keys %$args)
    {
        my $setQuery = $args->{$key};
        my $val = $self->_evalSet($setQuery, $self->_peekContext());
        $answers->{$key} = $val;
    }
    
    for my $key (keys %$args)
    {
        $self->_envSetSet($key, $answers->{$key});
    }
    
    return 1;
}

sub cmdSetVals
{
    my $self = shift;
    $self->_addAction('setval', $self->_parseHashAsValues(shift));
}

sub procSetVals
{
    my $self = shift;
    my ($args) = @_;

    # in order to avoid ugly behavior, we calculate all the answers, THEN
    #  we apply them to the env.
    my $answers = {};

    for my $key (keys %$args)
    {
        my $valQuery = $args->{$key};
        my $val = $self->_evalValue($valQuery, $self->_peekContext());
        $answers->{$key} = $val;
    }
    
    for my $key (keys %$args)
    {
        $self->_envSetVal($key, $answers->{$key});
    }
    
    return 1;
}

sub procLoopEnter
{
    my $self = shift;
    my ($args) = @_;
    
    if(!exists($args->{query}))
    {
        $self->_logErr("Loops require a 'query' to be specified." .
                       " Acting as if loop does not exist.");
        return;
    }
    
    my $ctxMgr = $self->_evalSet($args->{query}, $self->_peekContext());
    
    if(!defined($ctxMgr))
    {
        $self->_logErr("Query: " . $args->{query} . " did not return a " .
                        "context.");
        return;
    }
    
    if(exists($args->{name}))
    {
        # save it in the environment
        $self->_envSetSet($args->{name}, $ctxMgr);
    }
    
    $self->_pushContext($ctxMgr);
}

sub procLoopExit
{
    my $self = shift;
    my ($args) = @_;
    
    if(!exists($args->{query}))
    {
        # we're ignoring this loop. do no further processing.
        return;
    }
    
    $self->_popContext();
}

sub procLoopBody
{
    my $self = shift;
    my ($args, $actionList) = @_;

    my $ctxMgr = $self->_peekContext();
    
#    use Data::Dump qw(dump);
#    warn 'Contexts: ' . dump($self->{CONTEXT_STACK});
    
    $ctxMgr->loop(sub
    {
        my ($testCtxMgr, $node) = @_;
        
        if($args->{name})
        {
            # it must be a value; there's no way to get a context manager full
            #  of context managers
            $self->_envSetVal($args->{name}, $node);
        }
        
        $self->_pushContext($testCtxMgr);
        
        $self->_runList($actionList);
        
        $self->_popContext();
    });
    
    1;
}

sub cmdLoop
{
    my $self = shift;
    my ($context, $args) = @_;

    my $builder = $context->stash('builder');
    my $tokens  = $context->stash('tokens');
    
    my $actionList = [];
           
    $self->_pushActionList($actionList);
    
    $builder->build($context, $tokens);
    
    $self->_popActionList();

    # name should be left intact, so just parse query inplace.
    $args->{query} = $self->_parseValue($args->{query});    
    
    $self->_addAction('loop', [$args, $actionList]);
    
    1;
}

sub procLoop
{
    my $self = shift;
    my ($args, $actionList) = @{scalar shift};

    $self->procLoopEnter($args);

    $self->procLoopBody($args, $actionList);
    
    $self->procLoopExit($args);
}

sub cmdIf
{
    my $self = shift;
    my ($context, $args) = @_;

    $args = $self->_parseHashAsBooleans($args);

    $self->{ELSE_CASE} = undef;

    my $builder = $context->stash('builder');
    my $tokens  = $context->stash('tokens');
   
    my $actionList = [];
           
    $self->_pushActionList($actionList);
    
    $builder->build($context, $tokens);
    
    $self->_popActionList();
    
    $self->_addAction('if', [$args, $actionList, $self->{ELSE_CASE}]);
    
    1;
}

sub cmdElse
{
    my $self = shift;
    my ($context, $args) = @_;

    my $builder = $context->stash('builder');
    my $tokens  = $context->stash('tokens');
    
    my $actionList = [];
           
    $self->_pushActionList($actionList);
    
    $builder->build($context, $tokens);
    
    $self->_popActionList();
    
    $self->{ELSE_CASE} = $actionList;
    
    1;
}

sub procIf
{
    my $self = shift;
    my ($args, $trueActionList, $falseActionList) = @{scalar shift};
    
    if(exists $args->{test})
    {
        if($self->_evalBoolean($args->{test}, $self->_peekContext()))
        {
            $self->_runList($trueActionList);
        }
        elsif(defined($falseActionList))
        {
            $self->_runList($falseActionList);
        }
    }
    
    return 1;
}

sub transformElse
{
    my $self = shift;
    my $node = shift;
    
    if($node->[0] eq 'Else')
    {
        $node->[0] = 'KRVelse';
    }
    if((ref($node->[1]) eq 'HASH') and (scalar(@$node) > 2))
    {
        my $nodes = $node->[2];
        for my $curNode (@$nodes)
        {
            $self->transformElse($curNode);
        }
    }
    
}

sub procBody
{
    my $self = shift;
    my ($context, $args) = @_;

    my $builder = $context->stash('builder');
    my $tokens  = $context->stash('tokens');

    for my $token (@$tokens)
    {
        $self->transformElse($token);
    }

    $builder->build($context, $tokens);
    
    1;    
}

sub cmdFilter
{
    my $self = shift;
    my $args = shift;
    
    my $name = $args->{name};
    delete $args->{name};
    my $newArgs = $self->_parseHashAsValues($args);
    $newArgs->{name} = $name;
    
    $self->_addAction('filter', $newArgs);
}

sub procFilter
{
    my $self = shift;
    my ($args) = @_;
    my $drawing = $self->{DRAWING};

    # if not specified, the layer is 0
    my $filterArgs = {layer => 0};

    if(!exists $args->{name})
    {
        $self->_logErr('Filter requires a name!');
        return;
    }
    if(!$drawing->supportsFilter($args->{name}))
    {
        $self->_logErr('Filter ' . $args->{name} . ' is not supported.');
        return;
    }
    
    # copy it verbatim, no evaluation
    $filterArgs->{name} = $args->{name};

    for my $argKey (keys %$args)
    {
        my $argVal = $args->{$argKey};
    
        if($argKey ne 'name')
        {
            $filterArgs->{$argKey} =
                 $self->_evalValue($argVal, $self->_peekContext());
        }
    }

    $drawing->addFilter($filterArgs);
}

sub _procShapeCommon
{
    my $self = shift;
    my ($shape, $args) = @_;
    
    # process width/height/etc. BEFORE x,y, which are center and affected by
    #  width/height.  (center x,y not magically updating)
    if(exists $args->{edge})
    {
        my $edge = $self->_evalValue($args->{edge}, $self->_peekContext());
        $shape->width($edge);
        $shape->height($edge);
    }

    if(exists $args->{width})
    {
        my $width = $self->_evalValue($args->{width}, $self->_peekContext());
        $shape->width($width);
    }
    
    if(exists $args->{height})
    {
        my $height = $self->_evalValue($args->{height}, $self->_peekContext());
        $shape->height($height);
    }

    if(exists $args->{left})
    {
        my $left = $self->_evalValue($args->{left}, $self->_peekContext());
        $shape->left($left);
    }
    
    if(exists $args->{right})
    {
        my $right = $self->_evalValue($args->{right}, $self->_peekContext());
        $shape->right($right);
    }

    if(exists $args->{top})
    {
        my $top = $self->_evalValue($args->{top}, $self->_peekContext());
        $shape->left($top);
    }
    
    if(exists $args->{bottom})
    {
        my $bottom = $self->_evalValue($args->{bottom}, $self->_peekContext());
        $shape->right($bottom);
    }

    if(exists $args->{'x'})
    {
        my $x = $self->_evalValue($args->{'x'}, $self->_peekContext());
        $shape->centerX($x);
    }
    
    if(exists $args->{'y'})
    {
        my $y = $self->_evalValue($args->{'y'}, $self->_peekContext());
        $shape->centerY($y);
    }
    
    if(exists $args->{edgecolor})
    {
        my $color = $self->_evalValue($args->{edgecolor}, $self->_peekContext());
        $shape->edgeColor($color);
    }
    
    if(exists $args->{fillcolor})
    {
        my $color = $self->_evalValue($args->{fillcolor}, $self->_peekContext());
        $shape->fillColor($color);
    }
    
    if(exists $args->{layer})
    {
        my $layer = $self->_evalValue($args->{layer}, $self->_peekContext());
        $shape->layer($layer);
    }
    
    if(exists $args->{startang})
    {
        my $startang = $self->_evalValue($args->{startang}, $self->_peekContext());
        $shape->startang($startang);
    }
    
    if(exists $args->{endang})
    {
        my $endang = $self->_evalValue($args->{endang}, $self->_peekContext());
        $shape->endang($endang);
    }

    if(exists $args->{url})
    {
        my $url = $self->_evalValue($args->{url}, $self->_peekContext());
        $shape->url($url);
    }
}

sub cmdRect
{
    my $self = shift;
    $self->_addAction('rect', $self->_parseHashAsValues(shift));
}

sub procRect
{
    my $self = shift;
    my ($args) = @_;
    
    my $shape = OnlyTheWind::KoalaRainbow::DrawNodes::Rectangle->new();
    $self->_procShapeCommon($shape, $args);

    my $drawing = $self->{DRAWING};
    $drawing->addNode($shape);
}

sub cmdEllipse
{
    my $self = shift;
    $self->_addAction('ellipse', $self->_parseHashAsValues(shift));
}

sub procEllipse
{
    my $self = shift;
    my ($args) = @_;
    
    my $shape = OnlyTheWind::KoalaRainbow::DrawNodes::Ellipse->new();
    $self->_procShapeCommon($shape, $args);

    my $drawing = $self->{DRAWING};
    $drawing->addNode($shape);
}

sub cmdLine
{
    my $self = shift;
    $self->_addAction('line', $self->_parseHashAsValues(shift));
}

sub procLine
{
    my $self = shift;
    my ($args) = @_;
    
    my $shape = OnlyTheWind::KoalaRainbow::DrawNodes::Line->new();
    # not the best fit, but hey. just don't set unused things :)
    $self->_procShapeCommon($shape, $args);

    if(exists $args->{x1})
    {
        my $x1 = $self->_evalValue($args->{x1}, $self->_peekContext());
        $shape->x1($x1);
    }

    if(exists $args->{x2})
    {
        my $x2 = $self->_evalValue($args->{x2}, $self->_peekContext());
        $shape->x2($x2);
    }
    
    if(exists $args->{y1})
    {
        my $y1 = $self->_evalValue($args->{y1}, $self->_peekContext());
        $shape->y1($y1);
    }

    if(exists $args->{y2})
    {
        my $y2 = $self->_evalValue($args->{y2}, $self->_peekContext());
        $shape->y2($y2);
    }
    
    my $drawing = $self->{DRAWING};
    $drawing->addNode($shape);
}

1;
