#===========================================================================
# KoalaRainbow - a visualization system
# Copyright (C) 2004, Andrew Sutherland (sombrero@alum.mit.edu)
#
# KoalaRainbow is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#===========================================================================
package OnlyTheWind::KoalaRainbow::WebGui;
use strict;

use MT::App;
@OnlyTheWind::KoalaRainbow::WebGui::ISA = qw( MT::App );

use File::Spec;

use MT::Builder;
use MT::Template::Context;

use MT::Blog;

use OnlyTheWind::KoalaRainbow::QueryParser;
use OnlyTheWind::KoalaRainbow::ContextManager;

use File::Spec;

sub init
{
    my $self = shift;
    $self->SUPER::init(@_) or return;
    
    # Register our processing methods.
    $self->add_methods(
        'menu' => \&menuMode,
        'queries' => \&queriesMode,
        'visualizations' => \&visualizationMode
        );
    
    # start at the main menu
    $self->{default_mode} = 'menu';

    # note that we do some magic in load_tmpl to provide failover support for
    #  includes to the cms directory, and to route login.tmpl requests there
    #  as well.    
    $self->{template_dir} = 'koalaRainbow';
    
    # Only people with an account can login.  We will further restrict this
    #  so that they must have template editing permissions can use us.
    $self->{requires_login} = 1;
    
    # Use the standard MT user authentication mechanism; author!
    $self->{user_class} = 'MT::Author';
    
    # Overwrite the default breadcrumb
    $self->{breadcrumbs} = [ { bc_name => $self->translate('Koala Rainbow'),
                               bc_uri => $self->plugin_uri() } ];
    
    $self;
}

# Authorize user.  The user must have template editing permissions on some blog
#  in able to use our functionality.  The rationale is that we allow arbitrary
#  queries to be run against the entire database, so the user had better be able
#  to already get at all this information; also, it only makes sense to be able
#  to use this interface if the user can actually modify templates to use our
#  functionality in production.  The template editing permission meets both of
#  these requirements.  If the user can change templates, the user can have
#  any information in the database spit out into an html file (from any blog),
#  as I do not  believe there are any protections against this in the system.
#  Likewise, KR is only useful to those who can invoke its mechanisms in
#  templates.
sub is_authorized
{
    my $self = shift;
    require MT::Permission;
    
    if( $self->{author} )
    {
        my @perms  = MT::Permission->load({author_id => $self->{author}->id });
            
        for my $perm (@perms)
        {
            if($perm->can_edit_templates)
            {
                $self->{AUTH_BLOG_ID} = $perm->blog_id;
                return 1;
            }
        }
        
        $self->error(
            $self->translate("You are not authorized to use this tool."));
    }
    else # is this even possible?  better redundant than rooted!
    {
        $self->error($self->translate("You need to log in!"));
    }
}

# hackety hack, don't hack back.
sub uri
{
    my $self = shift;
    
    if($self->{author})
    {
        return $self->path . MT::ConfigMgr->instance->AdminScript;
    }
    else
    {
        return $self->plugin_uri;
    }
}


# be tricky and add the cms template directory as a failover path.  Oh so
#  tricky! Muahaha!
sub load_tmpl
{
    my $self = shift;
    my($file, @p) = @_;
    
    # override path.
    push(@p, 'path');
    push(@p, [File::Spec->catdir($self->{cfg}->TemplatePath,
                                  'koalaRainbow'),
#                                 $self->{template_dir}),
              File::Spec->catfile($self->{cfg}->TemplatePath, 'cms')]);
    
    # gah. okay, so this now lets us handle direct logins properly
    # hackity hack, hack hack hack hack hack hack
    if($file eq 'login.tmpl')
    {
        $file = '../cms/login.tmpl';
    }    
    
    my $tmpl = $self->SUPER::load_tmpl($file, @p);
    
    #my $cmspath = ;
    
    #die join(' ', @{$tmpl->{options}{path}}) . $cmspath;
    #push(@{$tmpl->{options}{path}}, $cmspath);
    
    #use Carp;
    #confess 'foo';
    
    return $tmpl;
}

sub menuMode
{
    my $self = shift;
    my %params = (no_breadcrumbs => 1);
    
    $self->build_page('menu.tmpl', \%params);
}

sub _procQuery
{
    my $self = shift;
    my ($query, $params, $displayResults) = @_;
    
    my $tree;
    my $resCtxMgr;
 
    # it comes in an array.  why?
    my $queryText = $query->{query_text}->[0];
    $queryText =~ tr/\r//d;
 
    my $context = MT::Template::Context->new();

    # use the blog we authenticated against
    my $blog_id = $self->{AUTH_BLOG_ID};
    $context->stash('blog_id', $blog_id);
    # get the blog object.
    my $blog = MT::Blog->load($blog_id);
    $context->stash('blog', $blog); 
 
    my $parser = OnlyTheWind::KoalaRainbow::QueryParser->new();
 
    eval
    {
        local $SIG{'__DIE__'};
        $tree = $parser->parseSetQuery($queryText);
    };
 
    # problem?
    if($@)
    {
        $params->{parse_error} = $@;
    }
    else # is okay
    {
        my $nullCtxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new();
        $nullCtxMgr->hint('mtcontext', $context);
     
        eval
        {
            local $SIG{'__DIE__'};
            
            $tree = $tree->partialEvaluate($nullCtxMgr);
            
            $resCtxMgr = $tree->evaluate($nullCtxMgr);
        };
        
        if($@)
        {
            $params->{eval_error} = $@;
        }
        elsif($displayResults)
        {
            eval '
            use Data::Dump qw(dump);
            
            if(ref($resCtxMgr) eq "OnlyTheWind::KoalaRainbow::ContextManager")
            {
                $params->{query_results} = dump($resCtxMgr->nodes());
            }
            elsif(defined($resCtxMgr))
            {
                $params->{query_results} = dump($resCtxMgr);
            }
            else
            {
                $params->{query_results} = "There is a distinct lack of " .
                    "results.";
                
            }';
            if($@)
            {
                $params->{query_results} = 'Data::Dump is not installed, ' .
                    'cannot print query outputs.';
            }
        }
    }
 
    $params->{query_text} = $queryText;
    
    return $resCtxMgr;
}

sub queriesMode
{
    my $self = shift;
    my $query = $self->{query};
    
    my %params = ();
    
    if($query->{query_text})
    {
        $self->_procQuery($query, \%params, 1);
    }
    
    $self->add_breadcrumb($self->translate('Queries'));
    
    $self->build_page('queries.tmpl', \%params);
}

sub _procVis
{
    my $self = shift;
    my ($query, $params, $ctxMgr) = @_;

    my $visText = $query->{vis_text}->[0];
    $visText =~ tr/\r//d;

    my $builder = MT::Builder->new();
    my $context = MT::Template::Context->new();

    # use the blog we authenticated against
    my $blog_id = $self->{AUTH_BLOG_ID};
    $context->stash('blog_id', $blog_id);
    # get the blog object.
    my $blog = MT::Blog->load($blog_id);
    $context->stash('blog', $blog);
    
    $context->stash('entries', $ctxMgr->nodes);
    
    my $outputFilePath = File::Spec->catfile($blog->site_path, 'krtvis.png');
    my $outputURLPath = $blog->site_url;
    if(!$outputURLPath =~ m!/$!)
    {
        $outputURLPath .= '/';
    }
    $outputURLPath .= 'krtvis.png';
    $context->stash('KRoutputfilepathoverride', $outputFilePath);
    $context->stash('KRoutputurlpathoverride', $outputURLPath);

    $context->stash('KRsavevisobj', 1);

    my $tokens = $builder->compile($context, $visText);
    #if(!$tokens)
    {
        $params->{templ_error} = $builder->errstr();
    }
    
    my $startSecs = times;
    
    my $results = $builder->build($context, $tokens);

    my $endSecs = times;

    #if(!defined($results))
    {
        $params->{build_error} = $builder->errstr();
    }

    # get the vis object out, log errors
    my $vis = $context->stash('KRsavevisobj');
    if((ref($vis) eq 'OnlyTheWind::KoalaRainbow::Visualization') and
       $vis->errorCount())
    {
        $params->{vis_error} = $vis->errorMessages();
    }

    $params->{run_secs} = $endSecs - $startSecs;

    $params->{vis_text} = $visText;
    
    $params->{vis_rendering} = $results;
    
    1;
}

sub visualizationMode
{
    my $self = shift;
    my $query = $self->{query};
    
    my %params = ();
    
    my $baseCtxMgr;
    
    if($query->{query_text})
    {
        $baseCtxMgr = $self->_procQuery($query, \%params, 0);
    }
    else
    {
        $baseCtxMgr = OnlyTheWind::KoalaRainbow::ContextManager->new();
    }
    
    if($query->{vis_text})
    {
        $self->_procVis($query, \%params, $baseCtxMgr);
    }
    
    $self->add_breadcrumb($self->translate('Visualizations'));
    
    $self->build_page('visualizations.tmpl', \%params);
}

sub plugin_uri
{
    my $self = shift;
    
    # HACK, but at least one that won't break if the problem is fixed; the
    #  downside is we won't be flexible to our plugin being moved.  curses.
    # More specifically:
    # - fixed, moved deeper in plugins => okay
    # - fixed, moved out of plugins => breakage
    # - not fixed, moved anywhere => breakage
    if($self->path =~ m!plugins/! or $self->script =~ m!plugins/!)
    {
        return $self->path . $self->script;
    }
    else
    {
        return $self->path . 'plugins/' . $self->script;
    }
}

sub build_page
{
    my ($self, $templateName, $params) = @_;
    
    $params->{kr_version} = $OnlyTheWind::KoalaRainbow::KoalaRainbow::VERSION;
    $params->{plugin_url} = $self->plugin_uri();
    
    $self->SUPER::build_page($templateName, $params);
}

1;
